__all__ = ["ObjectDetectionRecord", "InstanceSegmentationRecord"]

from icevision.core.record import *
from icevision.core.record_components import *


def ObjectDetectionRecord():
    return BaseRecord(
        (
            FilepathRecordComponent(),
            InstancesLabelsRecordComponent(),
            BBoxesRecordComponent(),
        )
    )


def InstanceSegmentationRecord():
    return BaseRecord(
        (
            FilepathRecordComponent(),
            InstancesLabelsRecordComponent(),
            BBoxesRecordComponent(),
            MasksRecordComponent(),
        )
    )



__all__ = [
    "Adapter",
    "AlbumentationsAdapterComponent",
    "AlbumentationsImgComponent",
    "AlbumentationsSizeComponent",
    "AlbumentationsInstancesLabelsComponent",
    "AlbumentationsBBoxesComponent",
    "AlbumentationsMasksComponent",
    "AlbumentationsKeypointsComponent",
    "AlbumentationsIsCrowdsComponent",
]

import albumentations as A
from itertools import chain

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.tfms.transform import *


@dataclass
class CollectOp:
    fn: Callable
    order: float = 0.5


class AlbumentationsAdapterComponent(Component):
    @property
    def adapter(self):
        return self.composite

    def setup(self):
        return

    def prepare(self, record):
        pass

    def collect(self, record):
        pass


class AlbumentationsImgComponent(AlbumentationsAdapterComponent):
    def setup_img(self, record):
        self.adapter._albu_in["image"] = record.img

        self.adapter._collect_ops.append(CollectOp(self.collect))

    def collect(self, record):
        record.set_img(self.adapter._albu_out["image"])


class AlbumentationsSizeComponent(AlbumentationsAdapterComponent):
    order = 0.2

    def setup_size(self, record):
        self.adapter._collect_ops.append(CollectOp(self.collect, order=0.2))

    def collect(self, record) -> ImgSize:
        # return self._size_no_padding
        width, height = self.adapter._size_no_padding
        record.set_image_size(width=width, height=height)


class AlbumentationsInstancesLabelsComponent(AlbumentationsAdapterComponent):
    order = 0.1

    def set_labels(self, record, labels):
        # TODO HACK: Will not work for multitask, will fail silently
        record.detection.set_labels_by_id(labels)

    def setup_instances_labels(self, record_component):
        # TODO HACK: Will not work for multitask, will fail silently
        self._original_labels = record_component.label_ids
        # Substitue labels with list of idxs, so we can also filter out iscrowds in case any bboxes are removed
        self.adapter._albu_in["labels"] = list(range(len(self._original_labels)))

        self.adapter._collect_ops.append(CollectOp(self.collect_labels, order=0.1))

    def collect_labels(self, record):
        self.adapter._keep_mask = np.zeros(len(self._original_labels), dtype=bool)
        self.adapter._keep_mask[self.adapter._albu_out["labels"]] = True

        labels = self.adapter._filter_attribute(self._original_labels)
        self.set_labels(record, labels)


class AlbumentationsBBoxesComponent(AlbumentationsAdapterComponent):
    def setup_bboxes(self, record_component):
        self.adapter._compose_kwargs["bbox_params"] = A.BboxParams(
            format="pascal_voc", label_fields=["labels"]
        )
        # TODO: albumentations has a way of sending information that can be used for tasks

        # TODO HACK: Will not work for multitask, will fail silently
        self.adapter._albu_in["bboxes"] = [o.xyxy for o in record_component.bboxes]

        self.adapter._collect_ops.append(CollectOp(self.collect))

    def collect(self, record) -> List[BBox]:
        # TODO: quickfix from 576
        # bboxes_xyxy = [_clip_bboxes(xyxy, img_h, img_w) for xyxy in d["bboxes"]]
        bboxes_xyxy = [xyxy for xyxy in self.adapter._albu_out["bboxes"]]
        bboxes = [BBox.from_xyxy(*xyxy) for xyxy in bboxes_xyxy]
        # TODO HACK: Will not work for multitask, will fail silently
        record.detection.set_bboxes(bboxes)

    @staticmethod
    def _clip_bboxes(xyxy, h, w):
        """Clip bboxes coordinates that are outside image dimensions."""
        x1, y1, x2, y2 = xyxy
        if w >= h:
            pad = (w - h) // 2
            h1 = pad
            h2 = w - pad
            return (x1, max(y1, h1), x2, min(y2, h2))
        else:
            pad = (h - w) // 2
            w1 = pad
            w2 = h - pad
            return (max(x1, w1), y1, min(x2, w2), y2)


class AlbumentationsMasksComponent(AlbumentationsAdapterComponent):
    def setup_masks(self, record):
        self.adapter._albu_in["masks"] = list(record.masks.data)
        self.adapter._collect_ops.append(CollectOp(self.collect))

    def collect(self, record):
        masks = self.adapter._filter_attribute(self.adapter._albu_out["masks"])
        masks = MaskArray(np.array(masks))
        record.detection.set_masks(masks)


class AlbumentationsKeypointsComponent(AlbumentationsAdapterComponent):
    def setup_keypoints(self, record_component):
        self.adapter._compose_kwargs["keypoint_params"] = A.KeypointParams(
            format="xy", remove_invisible=False, label_fields=["keypoints_labels"]
        )

        # not compatible with some transforms
        flat_tfms_list_ = _flatten_tfms(self.adapter.tfms_list)
        if get_transform(flat_tfms_list_, "RandomSizedBBoxSafeCrop") is not None:
            raise RuntimeError("RandomSizedBBoxSafeCrop is not supported for keypoints")

        self._kpts = record_component.keypoints
        self._kpts_xy = [xy for o in self._kpts for xy in o.xy]
        self._kpts_labels = [label for o in self._kpts for label in o.metadata.labels]
        self._kpts_visible = [visible for o in self._kpts for visible in o.visible]
        assert len(self._kpts_xy) == len(self._kpts_labels) == len(self._kpts_visible)

        self.adapter._albu_in["keypoints"] = self._kpts_xy
        self.adapter._albu_in["keypoints_labels"] = self._kpts_labels

        self.adapter._collect_ops.append(CollectOp(self.collect))

    def collect(self, record):
        # remove_invisible=False, therefore all points getting in are also getting out
        assert len(self.adapter._albu_out["keypoints"]) == len(self._kpts_xy)

        tfmed_kpts = self._remove_albu_outside_keypoints(
            tfms_kpts=self.adapter._albu_out["keypoints"],
            kpts_visible=self._kpts_visible,
            size_no_padding=self.adapter._size_no_padding,
        )
        # flatten list of keypoints
        flat_kpts = list(chain.from_iterable(tfmed_kpts))
        # group keypoints from same instance
        group_kpts = [
            flat_kpts[i : i + len(flat_kpts) // len(self._kpts)]
            for i in range(
                0,
                len(flat_kpts),
                len(flat_kpts) // len(self._kpts),
            )
        ]
        assert len(group_kpts) == len(self._kpts)

        kpts = [
            KeyPoints.from_xyv(group_kpt, original_kpt.metadata)
            for group_kpt, original_kpt in zip(group_kpts, self._kpts)
        ]
        kpts = self.adapter._filter_attribute(kpts)
        record.detection.set_keypoints(kpts)

    @classmethod
    def _remove_albu_outside_keypoints(cls, tfms_kpts, kpts_visible, size_no_padding):
        """Remove keypoints that are outside image dimensions."""
        v = kpts_visible
        v_n = v.copy()
        tra_n = tfms_kpts.copy()
        for i in range(len(tfms_kpts)):
            if v[i] > 0:
                v_n[i] = cls._check_kps_coords(tfms_kpts[i], size_no_padding)
                if v_n[i] == 1:
                    v_n[i] = v[i]
            if v_n[i] == 0:
                tra_n[i] = (0, 0)
            tra_n[i] = (tra_n[i][0], tra_n[i][1], v_n[i])
        return tra_n

    @staticmethod
    def _check_kps_coords(p, size_no_padding):
        x, y = p
        w, h = size_no_padding
        if w >= h:
            pad = (w - h) // 2
            h1 = pad
            h2 = w - pad
            return int((x <= w) and (x >= 0) and (y >= h1) and (y <= h2))
        else:
            pad = (h - w) // 2
            w1 = pad
            w2 = h - pad
            return int((x <= w2) and (x >= w1) and (y >= 0) and (y <= h))


class AlbumentationsIsCrowdsComponent(AlbumentationsAdapterComponent):
    def setup_iscrowds(self, record_component):
        self._iscrowds = record_component.iscrowds
        self.adapter._collect_ops.append(CollectOp(self.collect))

    def collect(self, record):
        iscrowds = self.adapter._filter_attribute(self._iscrowds)
        record.detection.set_iscrowds(iscrowds)


class AlbumentationsAreasComponent(AlbumentationsAdapterComponent):
    def setup_areas(self, record_component):
        self._areas = record_component.areas
        self.adapter._collect_ops.append(CollectOp(self.collect))

    def collect(self, record):
        areas = self.adapter._filter_attribute(self._areas)
        record.detection.set_areas(areas)


class Adapter(Transform, Composite):
    base_components = {
        AlbumentationsImgComponent,
        AlbumentationsSizeComponent,
        AlbumentationsInstancesLabelsComponent,
        AlbumentationsBBoxesComponent,
        AlbumentationsMasksComponent,
        AlbumentationsIsCrowdsComponent,
        AlbumentationsAreasComponent,
        AlbumentationsKeypointsComponent,
    }

    def __init__(self, tfms):
        super().__init__()
        self.tfms_list = tfms

    def create_tfms(self):
        return A.Compose(self.tfms_list, **self._compose_kwargs)

    def apply(self, record):
        # setup
        self._compose_kwargs = {}
        self._keep_mask = None
        self._albu_in = {}
        self._collect_ops = []
        record.setup_transform(tfm=self)

        # TODO: composing every time
        tfms = self.create_tfms()
        # apply transform
        self._albu_out = tfms(**self._albu_in)

        # store additional info (might be used by components on `collect`)
        self._size_no_padding = self._get_size_without_padding(record)

        # collect results
        for collect_op in sorted(self._collect_ops, key=lambda x: x.order):
            collect_op.fn(record)

        return record

    # def apply(self, record):
    #     self.prepare(record)

    #     self._albu_out = self.tfms(**self._albu_in)

    #     # store additional info (might be used by components on `collect`)
    #     self._size_no_padding = self._get_size_without_padding(record)

    #     self.reduce_on_components("collect", record=record)

    #     return record

    def _filter_attribute(self, v: list):
        if self._keep_mask is None:
            return v
        assert len(v) == len(self._keep_mask)
        return [o for o, keep in zip(v, self._keep_mask) if keep]

    def _get_size_without_padding(self, record) -> ImgSize:
        height, width, _ = self._albu_out["image"].shape

        if get_transform(self.tfms_list, "Pad") is not None:
            after_pad_h, after_pad_w, _ = record.img.shape

            t = get_transform(self.tfms_list, "SmallestMaxSize")
            if t is not None:
                presize = t.max_size
                height, width = _func_max_size(after_pad_h, after_pad_w, presize, min)

            t = get_transform(self.tfms_list, "LongestMaxSize")
            if t is not None:
                size = t.max_size
                height, width = _func_max_size(after_pad_h, after_pad_w, size, max)

        return ImgSize(width=width, height=height)


def _flatten_tfms(t):
    flat = []
    for o in t:
        if _is_iter(o):
            flat += [i for i in o]
        else:
            flat.append(o)
    return flat


def _is_iter(o):
    try:
        i = iter(o)
        return True
    except:
        return False


def get_transform(tfms_list, t):
    for el in tfms_list:
        if t in str(type(el)):
            return el
    return None


def py3round(number):
    """Unified rounding in all python versions."""
    if abs(round(number) - number) == 0.5:
        return int(2.0 * round(number / 2.0))

    return int(round(number))


def _func_max_size(height, width, max_size, func):
    scale = max_size / float(func(width, height))

    if scale != 1.0:
        height, width = tuple(py3round(dim * scale) for dim in (height, width))
    return height, width



__all__ = [
    "resnet18",
    "resnet34",
    "resnet50",
    "resnet101",
    "resnet152",
    "resnext101_32x8d",
]

from icevision.backbones import resnet
from icevision.models.torchvision.backbones.backbone_config import (
    TorchvisionBackboneConfig,
)

resnet18 = TorchvisionBackboneConfig(resnet.resnet18)
resnet34 = TorchvisionBackboneConfig(resnet.resnet34)
resnet50 = TorchvisionBackboneConfig(resnet.resnet50)
resnet101 = TorchvisionBackboneConfig(resnet.resnet101)
resnet152 = TorchvisionBackboneConfig(resnet.resnet152)
resnext101_32x8d = TorchvisionBackboneConfig(resnet.resnext101_32x8d)



# object detection
from icevision.models.mmdet.models import faster_rcnn
from icevision.models.mmdet.models import retinanet
from icevision.models.mmdet.models import fcos
from icevision.models.mmdet.models import cornernet
from icevision.models.mmdet.models import centripetalnet
from icevision.models.mmdet.models import sparse_rcnn

# segmentation
from icevision.models.mmdet.models import mask_rcnn



__all__ = [
    "build_train_batch",
    "build_valid_batch",
    "build_infer_batch",
    "train_dl",
    "valid_dl",
    "infer_dl",
]

from icevision.models.torchvision.faster_rcnn.dataloaders import *



__all__ = [
    "ImgSize",
    "open_img",
    "get_image_size",
    "get_img_size",
    "show_img",
    "plot_grid",
]

from icevision.imports import *
import PIL, imagesize

ImgSize = namedtuple("ImgSize", "width,height")


def open_img(fn, gray=False):
    color = "L" if gray else "RGB"
    return np.array(PIL.Image.open(str(fn)).convert(color))


# TODO: Deprecated
def get_image_size(filepath: Union[str, Path]) -> Tuple[int, int]:
    """
    Returns image (width, height)
    """
    logger.warning("get_image_size is deprecated, use get_img_size instead")
    return imagesize.get(filepath)


def get_img_size(filepath: Union[str, Path]) -> ImgSize:
    """
    Returns image (width, height)
    """
    return ImgSize(*imagesize.get(filepath))


def show_img(img, ax=None, show: bool = False, **kwargs):
    img = img.squeeze().copy()
    cmap = "gray" if len(img.shape) == 2 else None

    if ax is None:
        fig, ax = plt.subplots(**kwargs)

    ax.imshow(img, cmap=cmap)
    ax.set_axis_off()

    if show:
        plt.show()

    return ax


def plot_grid(
    fs: List[callable], ncols=1, figsize=None, show=False, axs_per_iter=1, **kwargs
):
    nrows = math.ceil(len(fs) * axs_per_iter / ncols)
    figsize = figsize or (12 * ncols, 12 * nrows)

    fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize, **kwargs)
    axs = np.asarray(axs)

    if axs_per_iter == 1:
        axs = axs.flatten()
    elif axs_per_iter > 1:
        axs = axs.reshape(-1, axs_per_iter)
    else:
        raise ValueError("axs_per_iter has to be greater than 1")

    for f, ax in zip(fs, axs):
        f(ax=ax)

    plt.tight_layout()
    if show:
        plt.show()



__all__ = [
    "build_train_batch",
    "build_valid_batch",
    "build_infer_batch",
    "train_dl",
    "valid_dl",
    "infer_dl",
]

from icevision.imports import *
from icevision.models.utils import *


def train_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    """A `DataLoader` with a custom `collate_fn` that batches items as required for training the model.

    # Arguments
        dataset: Possibly a `Dataset` object, but more generally, any `Sequence` that returns records.
        batch_tfms: Transforms to be applied at the batch level.
        **dataloader_kwargs: Keyword arguments that will be internally passed to a Pytorch `DataLoader`.
        The parameter `collate_fn` is already defined internally and cannot be passed here.

    # Returns
        A Pytorch `DataLoader`.
    """
    return transform_dl(
        dataset=dataset,
        build_batch=build_train_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def valid_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    """A `DataLoader` with a custom `collate_fn` that batches items as required for validating the model.

    # Arguments
        dataset: Possibly a `Dataset` object, but more generally, any `Sequence` that returns records.
        batch_tfms: Transforms to be applied at the batch level.
        **dataloader_kwargs: Keyword arguments that will be internally passed to a Pytorch `DataLoader`.
        The parameter `collate_fn` is already defined internally and cannot be passed here.

    # Returns
        A Pytorch `DataLoader`.
    """
    return transform_dl(
        dataset=dataset,
        build_batch=build_valid_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def infer_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    """A `DataLoader` with a custom `collate_fn` that batches items as required for inferring the model.

    # Arguments
        dataset: Possibly a `Dataset` object, but more generally, any `Sequence` that returns records.
        batch_tfms: Transforms to be applied at the batch level.
        **dataloader_kwargs: Keyword arguments that will be internally passed to a Pytorch `DataLoader`.
        The parameter `collate_fn` is already defined internally and cannot be passed here.

    # Returns
        A Pytorch `DataLoader`.
    """
    return transform_dl(
        dataset=dataset,
        build_batch=build_infer_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def build_train_batch(records, batch_tfms=None):
    """Builds a batch in the format required by the model when training.

    # Arguments
        records: A `Sequence` of records.
        batch_tfms: Transforms to be applied at the batch level.

    # Returns
        A tuple with two items. The first will be a tuple like `(images, targets)`,
        in the input format required by the model. The second will be an updated list
        of the input records with `batch_tfms` applied.

    # Examples

    Use the result of this function to feed the model.
    ```python
    batch, records = build_train_batch(records)
    outs = model(*batch)
    ```
    """
    records = common_build_batch(records, batch_tfms=batch_tfms)
    batch_images, batch_bboxes, batch_classes = zip(
        *(process_train_record(record) for record in records)
    )

    # convert to tensors
    batch_images = torch.stack(batch_images)
    batch_bboxes = [tensor(bboxes, dtype=torch.float32) for bboxes in batch_bboxes]
    batch_classes = [tensor(classes, dtype=torch.float32) for classes in batch_classes]

    # convert to EffDet interface
    targets = dict(bbox=batch_bboxes, cls=batch_classes)

    return (batch_images, targets), records


def build_valid_batch(records, batch_tfms=None):
    """Builds a batch in the format required by the model when validating.

    # Arguments
        records: A `Sequence` of records.
        batch_tfms: Transforms to be applied at the batch level.

    # Returns
        A tuple with two items. The first will be a tuple like `(batch_images, targets)`,
        in the input format required by the model. The second will be an updated list
        of the input records with `batch_tfms` applied.

    # Examples

    Use the result of this function to feed the model.
    ```python
    batch, records = build_valid_batch(records)
    outs = model(*batch)
    ```
    """
    (batch_images, targets), records = build_train_batch(records, batch_tfms)

    # convert to EffDet interface, when not training, dummy size and scale is required
    targets = dict(img_size=None, img_scale=None, **targets)

    return (batch_images, targets), records


def build_infer_batch(records, batch_tfms=None):
    """Builds a batch in the format required by the model when doing inference.

    # Arguments
        records: A `Sequence` of records.
        batch_tfms: Transforms to be applied at the batch level.

    # Returns
        A tuple with two items. The first will be a tuple like `(images, targets)`,
        in the input format required by the model. The second will be an updated list
        of the input records with `batch_tfms` applied. # Examples
    Use the result of this function to feed the model.
    ```python
    batch, records = build_infer_batch(records)
    outs = model(*batch)
    ```
    """
    records = common_build_batch(records, batch_tfms=batch_tfms)
    batch_images, batch_sizes, batch_scales = zip(
        *(process_infer_record(record) for record in records)
    )

    # convert to tensors
    batch_images = torch.stack(batch_images)
    batch_sizes = tensor(batch_sizes, dtype=torch.float32)
    batch_scales = tensor(batch_scales, dtype=torch.float32)

    # convert to EffDet interface
    targets = dict(img_size=batch_sizes, img_scale=batch_scales)

    return (batch_images, targets), records


def process_train_record(record) -> tuple:
    """Extracts information from record and prepares a format required by the EffDet training"""
    image = im2tensor(record.img)
    # background and dummy if no label in record
    classes = record.detection.label_ids if record.detection.label_ids else [0]
    bboxes = (
        [bbox.yxyx for bbox in record.detection.bboxes]
        if len(record.detection.label_ids) > 0
        else [[0, 0, 0, 0]]
    )
    return image, bboxes, classes


def process_infer_record(record) -> tuple:
    """Extracts information from record and prepares a format required by the EffDet inference"""
    image = im2tensor(record.img)
    image_size = image.shape[-2:]
    image_scale = 1.0

    return image, image_size, image_scale



from icevision.backbones.backbone_config import *
from icevision.backbones.vgg import *
from icevision.backbones.resnet import *
from icevision.backbones.resnet_fpn import *
from icevision.backbones.resnet_fpn_utils import *
from icevision.backbones.resnest_fpn import *
from icevision.backbones.mobilenet import *

# TODO: Deprecate
import icevision.backbones.resnet_fpn
import icevision.backbones.resnest_fpn



__all__ = [
    "coco",
    "COCOBaseParser",
    "COCOBBoxParser",
    "COCOMaskParser",
    "COCOKeyPointsParser",
    "COCOKeypointsMetadata",
]

from icevision.imports import *
from icevision.core import *
from icevision.utils import *
from icevision.parsers import *


# TODO: Deprecated
def coco(
    annotations_file: Union[str, Path],
    img_dir: Union[str, Path],
    mask: bool = True,
    idmap: Optional[IDMap] = None,
) -> Parser:
    logger.warning(
        "This function will be deprecated, instantiate the concrete "
        "classes instead: `COCOBBoxParser`, `COCOMaskParser`, `COCOKeypointsParser`"
    )
    parser_cls = COCOMaskParser if mask else COCOBBoxParser
    return parser_cls(annotations_file, img_dir, idmap=idmap)


class COCOBaseParser(Parser):
    def __init__(
        self,
        annotations_filepath: Union[str, Path],
        img_dir: Union[str, Path],
        idmap: Optional[IDMap] = None,
    ):

        self.annotations_dict = json.loads(Path(annotations_filepath).read_bytes())
        self.img_dir = Path(img_dir)

        self._record_id2info = {o["id"]: o for o in self.annotations_dict["images"]}

        categories = self.annotations_dict["categories"]
        id2class = {o["id"]: o["name"] for o in categories}
        id2class[0] = BACKGROUND
        # coco has non sequential ids, we fill the blanks with `None`, check #668 for more info
        classes = [None for _ in range(max(id2class.keys()) + 1)]
        for i, name in id2class.items():
            classes[i] = name
        self.class_map = ClassMap(classes)

        super().__init__(template_record=self.template_record(), idmap=idmap)

    def __iter__(self):
        yield from self.annotations_dict["annotations"]

    def __len__(self):
        return len(self.annotations_dict["annotations"])

    def template_record(self) -> BaseRecord:
        return BaseRecord(
            (
                FilepathRecordComponent(),
                InstancesLabelsRecordComponent(),
                AreasRecordComponent(),
                IsCrowdsRecordComponent(),
            )
        )

    def prepare(self, o):
        self._info = self._record_id2info[o["image_id"]]

    def record_id(self, o) -> int:
        return o["image_id"]

    def filepath(self, o) -> Path:
        return self.img_dir / self._info["file_name"]

    def img_size(self, o) -> ImgSize:
        return get_img_size(self.filepath(o))

    def labels_ids(self, o) -> List[Hashable]:
        return [o["category_id"]]

    def areas(self, o) -> List[float]:
        return [o["area"]]

    def iscrowds(self, o) -> List[bool]:
        return [o["iscrowd"]]

    def parse_fields(self, o, record, is_new):
        if is_new:
            record.set_filepath(self.filepath(o))
            record.set_img_size(self.img_size(o))

        # TODO: is class_map still a issue here?
        record.detection.set_class_map(self.class_map)
        record.detection.add_labels_by_id(self.labels_ids(o))
        record.detection.add_areas(self.areas(o))
        record.detection.add_iscrowds(self.iscrowds(o))


class COCOBBoxParser(COCOBaseParser):
    def bboxes(self, o) -> List[BBox]:
        return [BBox.from_xywh(*o["bbox"])]

    def template_record(self) -> BaseRecord:
        record = super().template_record()
        record.add_component(BBoxesRecordComponent())
        return record

    def parse_fields(self, o, record, is_new):
        super().parse_fields(o, record, is_new=is_new)
        record.detection.add_bboxes(self.bboxes(o))


class COCOMaskParser(COCOBBoxParser):
    def masks(self, o) -> List[MaskArray]:
        seg = o["segmentation"]
        if o["iscrowd"]:
            return [RLE.from_coco(seg["counts"])]
        else:
            return [Polygon(seg)]

    def template_record(self) -> BaseRecord:
        record = super().template_record()
        record.add_component(MasksRecordComponent())
        return record

    def parse_fields(self, o, record, is_new):
        super().parse_fields(o, record, is_new=is_new)
        record.detection.add_masks(self.masks(o))


class COCOKeyPointsParser(COCOBBoxParser):
    def template_record(self) -> BaseRecord:
        record = super().template_record()
        record.add_component(KeyPointsRecordComponent())
        return record

    def keypoints(self, o) -> List[KeyPoints]:
        return (
            [KeyPoints.from_xyv(o["keypoints"], COCOKeypointsMetadata)]
            if sum(o["keypoints"]) > 0
            else []
        )

    def labels_ids(self, o) -> List[Hashable]:
        if sum(o["keypoints"]) <= 0:
            return []
        return super().labels_ids(o)

    def areas(self, o) -> List[float]:
        return [o["area"]] if sum(o["keypoints"]) > 0 else []

    def iscrowds(self, o) -> List[bool]:
        return [o["iscrowd"]] if sum(o["keypoints"]) > 0 else []

    def bboxes(self, o) -> List[BBox]:
        return [BBox.from_xywh(*o["bbox"])] if sum(o["keypoints"]) > 0 else []

    def parse_fields(self, o, record, is_new):
        super().parse_fields(o, record, is_new=is_new)
        record.detection.add_keypoints(self.keypoints(o))


class COCOConnectionsColor:
    head = (220, 30, 200)
    torso = (100, 240, 100)
    right_arm = (170, 170, 100)
    left_arm = (120, 120, 230)
    left_leg = (230, 120, 130)
    right_leg = (230, 180, 190)


class COCOKeypointsMetadata(KeypointsMetadata):
    labels = (
        "nose",
        "left_eye",
        "right_eye",
        "left_ear",
        "right_ear",
        "left_shoulder",
        "right_shoulder",
        "left_elbow",
        "right_elbow",
        "left_wrist",
        "right_wrist",
        "left_hip",
        "right_hip",
        "left_knee",
        "right_knee",
        "left_ankle",
        "right_ankle",
    )

    connections = (
        KeypointConnection(0, 1, COCOConnectionsColor.head),
        KeypointConnection(0, 2, COCOConnectionsColor.head),
        KeypointConnection(1, 2, COCOConnectionsColor.head),
        KeypointConnection(1, 3, COCOConnectionsColor.head),
        KeypointConnection(2, 4, COCOConnectionsColor.head),
        KeypointConnection(3, 5, COCOConnectionsColor.head),
        KeypointConnection(4, 6, COCOConnectionsColor.head),
        KeypointConnection(5, 6, COCOConnectionsColor.torso),
        KeypointConnection(5, 7, COCOConnectionsColor.left_arm),
        KeypointConnection(5, 11, COCOConnectionsColor.torso),
        KeypointConnection(6, 8, COCOConnectionsColor.right_arm),
        KeypointConnection(6, 12, COCOConnectionsColor.torso),
        KeypointConnection(7, 9, COCOConnectionsColor.left_arm),
        KeypointConnection(8, 10, COCOConnectionsColor.right_arm),
        KeypointConnection(11, 12, COCOConnectionsColor.torso),
        KeypointConnection(13, 11, COCOConnectionsColor.left_leg),
        KeypointConnection(14, 12, COCOConnectionsColor.right_leg),
        KeypointConnection(15, 13, COCOConnectionsColor.left_leg),
        KeypointConnection(16, 14, COCOConnectionsColor.right_leg),
    )



__all__ = ["logger_default_config", "ReplaySink"]

from icevision.imports import *

logger.level("AUTOFIX", 25)
logger.level("AUTOFIX-START", 25, color="<fg #1f6f8b><bold>")
logger.level("AUTOFIX-SUCCESS", 25, color="<green>")
logger.level("AUTOFIX-FAIL", 25, color="<fg #ea2c62>")
logger.level("AUTOFIX-REPORT", 25, color="<fg #9a7d0a>")


def logger_default_config():
    logger.remove()
    logger.add(
        sys.stderr,
        format="<level><bold>{level: <8}</></> - <level>{message}</> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan>",
        level="INFO",
        colorize=True,
    )


logger_default_config()


class ReplaySink:
    """Capture messages and replays them after leaving the block.

    # Examples
    ```python
    pre_replay = lambda: logger.info("something will happen")
    post_replay = lambda: logger.info("something did happen")
    with ReplaySink(pre_replay=pre_replay, post_replay=post_replay) as sink:
        logger.info('captured message')
    ```
    """

    def __init__(
        self,
        pre_replay: Optional[callable] = None,
        post_replay: Optional[callable] = None,
    ):
        def _noop():
            pass

        self.captured = []
        self.pre_replay = pre_replay or _noop
        self.post_replay = post_replay or _noop

    def __enter__(self):
        logger.remove()
        logger.add(self)
        return self

    def write(self, message):
        self.captured.append(message)

    def __exit__(self, type, value, traceback):
        logger_default_config()
        if self.captured:
            self.pre_replay()

            for msg in self.captured:
                record = msg.record

                level = record["level"].name
                try:
                    logger.level(level)
                except ValueError:
                    # If the level 'name' is not registered, use its 'no' instead
                    level = record["level"].no

                patched = logger.patch(lambda r: r.update(record))
                patched.log(level, record["message"])

            self.post_replay()



__all__ = [
    "resnet18_fpn",
    "resnet34_fpn",
    "resnet50_fpn",
    "resnet101_fpn",
    "resnet152_fpn",
    "resnext50_32x4d_fpn",
    "resnext101_32x8d_fpn",
    "wide_resnet50_2_fpn",
    "wide_resnet101_2_fpn",
]

from icevision.imports import *
from icevision.utils import *
from icevision.backbones.resnet_fpn_utils import *
from torchvision.models.detection.backbone_utils import resnet_fpn_backbone


def _resnet_fpn(name: str, pretrained: bool = True, **kwargs):
    model = resnet_fpn_backbone(backbone_name=name, pretrained=pretrained, **kwargs)
    patch_param_groups(model)

    return model


def resnet18_fpn(pretrained: bool = True, **kwargs):
    return _resnet_fpn("resnet18", pretrained=pretrained, **kwargs)


def resnet34_fpn(pretrained: bool = True, **kwargs):
    return _resnet_fpn("resnet34", pretrained=pretrained, **kwargs)


def resnet50_fpn(pretrained: bool = True, **kwargs):
    return _resnet_fpn("resnet50", pretrained=pretrained, **kwargs)


def resnet101_fpn(pretrained: bool = True, **kwargs):
    return _resnet_fpn("resnet101", pretrained=pretrained, **kwargs)


def resnet152_fpn(pretrained: bool = True, **kwargs):
    return _resnet_fpn("resnet152", pretrained=pretrained, **kwargs)


def resnext50_32x4d_fpn(pretrained: bool = True, **kwargs):
    return _resnet_fpn("resnext50_32x4d", pretrained=pretrained, **kwargs)


def resnext101_32x8d_fpn(pretrained: bool = True, **kwargs):
    return _resnet_fpn("resnext101_32x8d", pretrained=pretrained, **kwargs)


def wide_resnet50_2_fpn(pretrained: bool = False, **kwargs):
    return _resnet_fpn("wide_resnet50_2", pretrained=pretrained, **kwargs)


def wide_resnet101_2_fpn(pretrained: bool = False, **kwargs):
    return _resnet_fpn("wide_resnet101_2", pretrained=pretrained, **kwargs)



__all__ = ["CodeLine", "CodeTemplate"]

from icevision.imports import *


@dataclass
class CodeLine:
    text: str
    indent: int


class CodeTemplate:
    def __init__(self, indent_spaces=4):
        self.lines: List[CodeLine] = []
        self.indent_str = " " * indent_spaces

    def add_line(self, line: str, indent=0):
        self.lines.append(CodeLine(line, indent))

    def add_lines(self, lines: Sequence[str], indent=0):
        for line in lines:
            self.add_line(line, indent)

    def display(self):
        for line in self.lines:
            print(f"{self.indent_str*line.indent}{line.text}")



__all__ = ["KeyPoints", "KeypointConnection", "KeypointsMetadata"]

from icevision.imports import *
from .exceptions import *


class KeypointConnection:
    def __init__(self, p1, p2, color=None):
        self.p1 = p1
        self.p2 = p2
        self.color = color or (100, 100, 100)


class KeypointsMetadata:
    labels: Tuple[str] = NotImplemented
    connections: Tuple[KeypointConnection] = None


class KeyPoints:
    def __init__(
        self, keypoints: Union[List[int], np.array], metadata: Type[KeypointsMetadata]
    ):
        self.keypoints = np.array(keypoints)
        self.x = self.keypoints[0::3]
        self.y = self.keypoints[1::3]
        self.visible = self.keypoints[2::3]
        self.xy = [(x, y) for x, y in zip(self.x, self.y)]
        self.n_visible_keypoints = (self.visible > 0).sum()
        self.xyv = [(x, y, v) for x, y, v in zip(self.x, self.y, self.visible)]
        self.metadata = metadata

    @classmethod
    def from_xyv(cls, keypoints, labels):
        return cls(keypoints, labels)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} "
            f"({self.n_visible_keypoints} visible keypoints)>"
        )

    def __eq__(self, other) -> bool:
        if isinstance(other, KeyPoints):
            return (
                np.all(self.keypoints == other.keypoints)
                and self.metadata == other.metadata
            )
        return False



__all__ = [
    "draw_label",
    "bbox_polygon",
    "draw_mask",
    "as_rgb_tuple",
    "get_default_font",
]

from icevision.imports import *
from icevision.utils import *
from matplotlib import patches
from PIL import Image, ImageFont, ImageDraw
import PIL


def draw_label(ax, x, y, name, color, fontsize=18):
    ax.text(
        x + 1,
        y - 2,
        name,
        fontsize=fontsize,
        color="white",
        va="bottom",
        bbox=dict(facecolor=color, edgecolor=color, pad=2, alpha=0.9),
    )


def bbox_polygon(bbox):
    bx, by, bw, bh = bbox.xywh
    poly = np.array([[bx, by], [bx, by + bh], [bx + bw, by + bh], [bx + bw, by]])
    return patches.Polygon(poly)


def draw_mask(ax, mask, color):
    color_mask = np.ones((*mask.shape, 3)) * color
    ax.imshow(np.dstack((color_mask, mask * 0.5)))
    ax.contour(mask, colors=[color_mask[0, 0, :]], alpha=0.4)


def as_rgb_tuple(x: Union[np.ndarray, tuple, list, str]) -> tuple:
    "Convert np RGB values -> tuple for PIL compatibility"
    if isinstance(x, (np.ndarray, tuple, list)):
        if not len(x) == 3:
            raise ValueError(f"Expected 3 (RGB) numbers, got {len(x)}")
        if isinstance(x, np.ndarray):
            return tuple(x.astype(np.int))
        elif isinstance(x, tuple):
            return x
        elif isinstance(x, list):
            return tuple(x)
    elif isinstance(x, str):
        return PIL.ImageColor.getrgb(x)
    else:
        raise ValueError(f"Expected {{np.ndarray|list|tuple}}, got {type(x)}")


def get_default_font() -> str:
    import requests

    font_dir = get_root_dir() / "fonts"
    font_dir.mkdir(exist_ok=True)

    font_file = font_dir / "SpaceGrotesk-Medium.ttf"
    if not font_file.exists():
        URL = "https://raw.githubusercontent.com/airctic/storage/master/SpaceGrotesk-Medium.ttf"
        logger.info(
            "Downloading default `.ttf` font file - SpaceGrotesk-Medium.ttf from {} to {}",
            URL,
            font_file,
        )
        font_file.write_bytes(requests.get(URL).content)
    return str(font_file)



__all__ = [
    "create_coco_api",
    "convert_records_to_coco_style",
    "convert_preds_to_coco_style",
    "convert_record_to_coco_annotations",
    "coco_api_from_records",
    "coco_api_from_preds",
    "create_coco_eval",
]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from pycocotools.coco import COCO
from pycocotools.cocoeval import COCOeval


def create_coco_api(coco_records) -> COCO:
    """Create COCO dataset api

    Args:
        coco_records: Records in coco style (use convert_records_to_coco_style to convert
        records to coco style.
    """
    coco_ds = COCO()
    coco_ds.dataset = coco_records
    coco_ds.createIndex()

    return coco_ds


def coco_api_from_preds(preds, show_pbar: bool = False) -> COCO:
    coco_preds = convert_preds_to_coco_style(preds, show_pbar=show_pbar)
    return create_coco_api(coco_preds)


def coco_api_from_records(records, show_pbar: bool = False) -> COCO:
    """Create pycocotools COCO dataset from records"""
    coco_records = convert_records_to_coco_style(records, show_pbar=show_pbar)
    return create_coco_api(coco_records=coco_records)


def create_coco_eval(
    records, preds, metric_type: str, show_pbar: bool = False
) -> COCOeval:
    assert len(records) == len(preds)

    for record, pred in zip(records, preds):
        pred.record_id = record.record_id
        pred.height = record.height
        pred.width = record.width
        # needs 'filepath' for mask `coco.py#418`
        pred.filepath = record.filepath

    target_ds = coco_api_from_records(records, show_pbar=show_pbar)
    pred_ds = coco_api_from_preds(preds, show_pbar=show_pbar)
    return COCOeval(target_ds, pred_ds, metric_type)


def convert_record_to_coco_image(record) -> dict:
    image = {}
    image["id"] = record.record_id
    image["file_name"] = Path(record.filepath).name
    image["width"] = record.width
    image["height"] = record.height
    return image


def convert_record_to_coco_annotations(record):
    annotations_dict = {
        "image_id": [],
        "category_id": [],
        "bbox": [],
        "area": [],
        "iscrowd": [],
    }
    # build annotations field
    for label in record.detection.label_ids:
        annotations_dict["image_id"].append(record.record_id)
        annotations_dict["category_id"].append(label)

    for bbox in record.detection.bboxes:
        annotations_dict["bbox"].append(list(bbox.xywh))

    if hasattr(record.detection, "areas"):
        for area in record.detection.areas:
            annotations_dict["area"].append(area)
    else:
        for bbox in record.detection.bboxes:
            annotations_dict["area"].append(bbox.area)

    # HACK: Because of prepare_record, mask should always be `MaskArray`,
    # maybe the for loop is not required?
    if hasattr(record.detection, "masks"):
        masks = record.detection.masks

        if isinstance(masks, MaskArray):
            masks = masks.to_erles(record.height, record.width)

        if isinstance(masks, EncodedRLEs):
            annotations_dict["segmentation"] = masks.erles

        else:
            raise RuntimeError(
                "masks are expected to be EncodedRLEs only, "
                "if you get this error please open an issue on github."
            )
            # annotations_dict["segmentation"] = []
            # for mask in record.detection.masks:
            #     if isinstance(mask, MaskArray):
            #         # HACK: see previous hack
            #         assert len(mask.shape) == 2
            #         mask2 = MaskArray(mask.data[None])
            #         rles = mask2.to_coco_rle(record.height, record.width)
            #         annotations_dict["segmentation"].extend(rles)
            #     elif isinstance(mask, Polygon):
            #         annotations_dict["segmentation"].append(mask.points)
            #     elif isinstance(mask, RLE):
            #         coco_rle = {
            #             "counts": mask.to_coco(),
            #             "size": [record.height, record.width],
            #         }
            #         annotations_dict["segmentation"].append(coco_rle)
            #     elif isinstance(mask, MaskFile):
            #         rles = mask.to_coco_rle(record.height, record.width)
            #         annotations_dict["segmentation"].extend(rles)
            #     elif isinstance(mask, EncodedRLEs):
            #         annotations_dict["segmentation"].append(mask.erles)
            #     else:
            #         msg = f"Mask type {type(mask)} unsupported"
            #         raise ValueError(msg)

    # TODO: is auto assigning a value for iscrowds dangerous (may hurt the metric value?)
    if not hasattr(record.detection, "iscrowds"):
        record.detection.iscrowds = [0] * len(record.detection.label_ids)
    for iscrowd in record.detection.iscrowds:
        annotations_dict["iscrowd"].append(iscrowd)

    if hasattr(record.detection, "scores"):
        annotations_dict["score"] = record.detection.scores

    return annotations_dict


def convert_preds_to_coco_style(preds, show_pbar: bool = False):
    return convert_records_to_coco_style(
        records=preds, images=True, categories=False, show_pbar=show_pbar
    )


def convert_records_to_coco_style(
    records,
    images: bool = True,
    annotations: bool = True,
    categories: bool = True,
    show_pbar: bool = True,
):
    """Converts records from library format to coco format.
    Inspired from: https://github.com/pytorch/vision/blob/master/references/detection/coco_utils.py#L146
    """
    images_ = []
    annotations_dict = defaultdict(list)

    for record in pbar(records, show=show_pbar):
        if images:
            image_ = convert_record_to_coco_image(record)
            images_.append(image_)

        if annotations:
            annotations_ = convert_record_to_coco_annotations(record)
            for k, v in annotations_.items():
                annotations_dict[k].extend(v)

    if annotations:
        annotations_dict = dict(annotations_dict)
        if not allequal([len(o) for o in annotations_dict.values()]):
            raise RuntimeError("Mismatch lenght of elements")

        # convert dict of lists to list of dicts
        annotations_ = []
        for i in range(len(annotations_dict["image_id"])):
            annotation = {k: v[i] for k, v in annotations_dict.items()}
            # annotations should be initialized starting at 1 (torchvision issue #1530)
            annotation["id"] = i + 1
            annotations_.append(annotation)

    if categories:
        categories_set = set(annotations_dict["category_id"])
        categories_ = [{"id": i} for i in categories_set]

    res = {}
    if images_:
        res["images"] = images_
    if annotations:
        res["annotations"] = annotations_
    if categories:
        res["categories"] = categories_

    return res



__all__ = ["model"]

from icevision.imports import *
from icevision.backbones import resnet_fpn
from icevision.models.torchvision.utils import *

from torchvision.models.detection.keypoint_rcnn import (
    keypointrcnn_resnet50_fpn,
    KeypointRCNNPredictor,
    KeypointRCNN,
)
from torchvision.models.detection.faster_rcnn import FastRCNNPredictor


def model(
    num_keypoints: int,
    num_classes: int = 2,
    backbone: Optional[nn.Module] = None,
    remove_internal_transforms: bool = True,
    **keypoint_rcnn_kwargs
) -> nn.Module:
    """KeypointRCNN model implemented by torchvision.

    # Arguments
        num_keypoints: Number of keypoints (e.g. 17 in case of COCO).
        num_classes: Number of classes (including background).
        backbone: Backbone model to use. Defaults to a resnet50_fpn model.
        remove_internal_transforms: The torchvision model internally applies transforms
        like resizing and normalization, but we already do this at the `Dataset` level,
        so it's safe to remove those internal transforms.
        **keypoint_rcnn_kwargs: Keyword arguments that internally are going to be passed to
        `torchvision.models.detection.keypoint_rcnn.KeypointRCNN`.

    # Returns
        A Pytorch `nn.Module`.
    """
    if backbone is None:
        model = keypointrcnn_resnet50_fpn(
            pretrained=True, pretrained_backbone=True, **keypoint_rcnn_kwargs
        )

        in_channels = model.roi_heads.keypoint_predictor.kps_score_lowres.in_channels
        model.roi_heads.keypoint_predictor = KeypointRCNNPredictor(
            in_channels, num_keypoints
        )

        if num_classes != 2:
            in_features = model.roi_heads.box_predictor.cls_score.in_features
            model.roi_heads.box_predictor = FastRCNNPredictor(in_features, num_classes)

        resnet_fpn.patch_param_groups(model.backbone)
    else:
        model = KeypointRCNN(
            backbone,
            num_classes=num_classes,
            num_keypoints=num_keypoints,
            **keypoint_rcnn_kwargs
        )

    patch_rcnn_param_groups(model=model)

    if remove_internal_transforms:
        remove_internal_model_transforms(model)

    return model



__all__ = ["BaseRecord", "autofix_records"]

from icevision.imports import *
from icevision.utils import *
from collections.abc import MutableMapping
from icevision.core import tasks
from icevision.core.exceptions import *
from icevision.core.components import *
from icevision.core.record_components import *


# TODO: MutableMapping because of backwards compatability
# TODO: Rename to Record
class BaseRecord(TaskComposite):
    base_components = {RecordIDRecordComponent, SizeRecordComponent}

    def as_dict(self) -> dict:
        return self.reduce_on_components("as_dict", reduction="update")

    def num_annotations(self) -> Dict[str, int]:
        return self.reduce_on_components("_num_annotations", reduction="update")

    def check_num_annotations(self):
        tasks_num_annotations = self.num_annotations()
        for task, num_annotations in tasks_num_annotations.items():
            if len(set(num_annotations.values())) > 1:
                msg = "\n".join(
                    [f"\t- {v} for {k}" for k, v in num_annotations.items()]
                )
                raise AutofixAbort(
                    "Number of items should be the same for each annotation type"
                    f", but got for task {task}:\n{msg}"
                )

    def autofix(self):
        self.check_num_annotations()

        tasks_success_dict = self.reduce_on_components("_autofix", reduction="update")

        for task_name, success_dict in tasks_success_dict.items():
            success_list = np.array(list(success_dict.values()))
            if len(success_list) == 0:
                continue
            keep_mask = reduce(np.logical_and, success_list)
            discard_idxs = np.where(keep_mask == False)[0]

            for i in discard_idxs:
                logger.log(
                    "AUTOFIX-REPORT",
                    "Removed annotation with index: {}, "
                    "for more info check the AUTOFIX-FAIL messages above",
                    i,
                )
                self.remove_annotation(task_name=task_name, i=i)

        return tasks_success_dict

    # TODO: Might have weird interaction with task_components
    def remove_annotation(self, i: int, task_name: str):
        self.reduce_on_task_components("_remove_annotation", task_name=task_name, i=i)

    def aggregate_objects(self):
        return self.reduce_on_components("_aggregate_objects", reduction="update")

    # Instead of copying here, copy outside?
    def load(self) -> "BaseRecord":
        record = deepcopy(self)
        record.reduce_on_components("_load")
        return record

    def unload(self):
        self.reduce_on_components("_unload")

    def setup_transform(self, tfm):
        self.reduce_on_components("setup_transform", tfm=tfm)

    def builder_template(self) -> List[str]:
        res = self.reduce_on_components("builder_template", reduction="extend").values()
        return [line for lines in res for line in lines]

    def __repr__(self) -> str:
        tasks_reprs = self.reduce_on_components("_repr", reduction="extend")

        def join_one(reprs):
            return "".join(f"\n\t- {o}" for o in reprs)

        reprs = [f"{task}: {join_one(reprs)}" for task, reprs in tasks_reprs.items()]
        repr = "\n".join(reprs)

        return f"{self.__class__.__name__}\n\n{repr}"

    # # backwards compatiblity: implemented method to behave like a dict
    # def __getitem__(self, key):
    #     return self.as_dict()[key]

    # def __setitem__(self, key, value):
    #     setattr(self, key, value)

    # def __delitem__(self, key):
    #     delattr(self, key)

    # def __iter__(self):
    #     yield from self.as_dict()

    # def __len__(self):
    #     return len(self.as_dict())


def autofix_records(records: Sequence[BaseRecord]) -> Sequence[BaseRecord]:
    keep_records = []
    for record in records:

        def _pre_replay():
            logger.log(
                "AUTOFIX-START",
                "️🔨  Autofixing record with record_id: {}  ️🔨",
                record.record_id,
            )

        with ReplaySink(_pre_replay) as sink:
            try:
                record.autofix()
                keep_records.append(record)
            except AutofixAbort as e:
                logger.warning(
                    "🚫 Record could not be autofixed and will be removed because: {}",
                    str(e),
                )

    return keep_records



__all__ = ["COCOMetric", "COCOMetricType"]

from icevision.imports import *
from icevision.utils import *
from icevision.data import *
from icevision.metrics.metric import *


class COCOMetricType(Enum):
    """Available options for `COCOMetric`."""

    bbox = "bbox"
    mask = "segm"
    keypoint = "keypoints"


class COCOMetric(Metric):
    """Wrapper around [cocoapi evaluator](https://github.com/cocodataset/cocoapi)

    Calculates average precision.

    # Arguments
        metric_type: Dependent on the task you're solving.
        print_summary: If `True`, prints a table with statistics.
        show_pbar: If `True` shows pbar when preparing the data for evaluation.
    """

    def __init__(
        self,
        metric_type: COCOMetricType = COCOMetricType.bbox,
        print_summary: bool = False,
        show_pbar: bool = False,
    ):
        self.metric_type = metric_type
        self.print_summary = print_summary
        self.show_pbar = show_pbar
        self._records, self._preds = [], []

    def _reset(self):
        self._records.clear()
        self._preds.clear()

    def accumulate(self, preds):
        # TODO, HACK: this modify all references of the record
        for pred in preds:
            pred.pred.unload()
            pred.ground_truth.unload()

            self._records.append(pred.ground_truth)
            self._preds.append(pred.pred)

    def finalize(self) -> Dict[str, float]:
        with CaptureStdout():
            coco_eval = create_coco_eval(
                records=self._records,
                preds=self._preds,
                metric_type=self.metric_type.value,
                show_pbar=self.show_pbar,
            )
            coco_eval.evaluate()
            coco_eval.accumulate()

        with CaptureStdout(propagate_stdout=self.print_summary):
            coco_eval.summarize()

        stats = coco_eval.stats
        logs = {
            "AP (IoU=0.50:0.95) area=all": stats[0],
            "AP (IoU=0.50) area=all": stats[1],
            "AP (IoU=0.75) area=all": stats[2],
            "AP (IoU=0.50:0.95) area=small": stats[3],
            "AP (IoU=0.50:0.95) area=medium": stats[4],
            "AP (IoU=0.50:0.95) area=large": stats[5],
            "AR (IoU=0.50:0.95) area=all maxDets=1": stats[6],
            "AR (IoU=0.50:0.95) area=all maxDets=10": stats[7],
            "AR (IoU=0.50:0.95) area=all maxDets=100": stats[8],
            "AR (IoU=0.50:0.95) area=small maxDets=100": stats[9],
            "AR (IoU=0.50:0.95) area=medium maxDets=100": stats[10],
            "AR (IoU=0.50:0.95) area=large maxDets=100": stats[11],
        }

        self._reset()
        return logs



__all__ = [
    "build_train_batch",
    "build_valid_batch",
    "build_infer_batch",
    "train_dl",
    "valid_dl",
    "infer_dl",
]

from icevision.core import *
from icevision.imports import *
from icevision.models.utils import *


def train_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    return transform_dl(
        dataset=dataset,
        build_batch=build_train_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def valid_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    return transform_dl(
        dataset=dataset,
        build_batch=build_valid_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def infer_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    """A `DataLoader` with a custom `collate_fn` that batches items as required for inferring the model.

    # Arguments
        dataset: Possibly a `Dataset` object, but more generally, any `Sequence` that returns records.
        batch_tfms: Transforms to be applied at the batch level.
        **dataloader_kwargs: Keyword arguments that will be internally passed to a Pytorch `DataLoader`.
        The parameter `collate_fn` is already defined internally and cannot be passed here.

    # Returns
        A Pytorch `DataLoader`.
    """
    return transform_dl(
        dataset=dataset,
        build_batch=build_infer_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def build_train_batch(
    records: Sequence[RecordType], batch_tfms=None
) -> Tuple[dict, List[Dict[str, torch.Tensor]]]:
    records = common_build_batch(records=records, batch_tfms=batch_tfms)

    images, labels, bboxes, img_metas = [], [], [], []
    for record in records:
        images.append(_img_tensor(record))
        img_metas.append(_img_meta(record))
        labels.append(_labels(record))
        bboxes.append(_bboxes(record))

    data = {
        "img": torch.stack(images),
        "img_metas": img_metas,
        "gt_labels": labels,
        "gt_bboxes": bboxes,
    }

    return data, records


def build_valid_batch(
    records: Sequence[RecordType], batch_tfms=None
) -> Tuple[dict, List[Dict[str, torch.Tensor]]]:
    return build_train_batch(records=records, batch_tfms=batch_tfms)


def build_infer_batch(records, batch_tfms=None):
    records = common_build_batch(records, batch_tfms=batch_tfms)

    imgs, img_metas = [], []
    for record in records:
        imgs.append(_img_tensor(record))
        img_metas.append(_img_meta(record))

    data = {
        "img": [torch.stack(imgs)],
        "img_metas": [img_metas],
    }

    return data, records


def _img_tensor(record):
    # convert from RGB to BGR
    img = record.img[:, :, ::-1].copy()
    return im2tensor(img)


def _img_meta(record):
    img_h, img_w, img_c = record.img.shape

    return {
        # height and width from sample is before padding
        "img_shape": (record.img_size.height, record.img_size.width, img_c),
        "pad_shape": (img_h, img_w, img_c),
        "scale_factor": np.ones(4),  # TODO: is scale factor correct?
    }


def _labels(record):
    if len(record.detection.label_ids) == 0:
        raise RuntimeError("Negative samples still needs to be implemented")
    else:
        # convert background from id 0 to last
        labels = tensor(record.detection.label_ids, dtype=torch.int64) - 1
        labels[labels == -1] = record.detection.class_map.num_classes - 1
        return labels


def _bboxes(record):
    if len(record.detection.label_ids) == 0:
        raise RuntimeError("Negative samples still needs to be implemented")
    else:
        xyxys = [bbox.xyxy for bbox in record.detection.bboxes]
        return tensor(xyxys, dtype=torch.float32)



__all__ = ["predict", "predict_dl", "convert_raw_predictions"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.data import *
from icevision.models.utils import _predict_dl
from icevision.models.ross.efficientdet.dataloaders import *
from effdet import DetBenchTrain, DetBenchPredict, unwrap_bench


@torch.no_grad()
def _predict_batch(
    model: Union[DetBenchTrain, DetBenchPredict],
    batch: Sequence[torch.Tensor],
    records: Sequence[BaseRecord],
    detection_threshold: float = 0.5,
    device: Optional[torch.device] = None,
) -> List[Prediction]:
    device = device or model_device(model)

    imgs, img_info = batch
    imgs = imgs.to(device)
    img_info = {k: v.to(device) for k, v in img_info.items()}

    bench = DetBenchPredict(unwrap_bench(model))
    bench = bench.eval().to(device)

    raw_preds = bench(x=imgs, img_info=img_info)
    preds = convert_raw_predictions(
        raw_preds=raw_preds, records=records, detection_threshold=detection_threshold
    )

    return preds


def predict(
    model: Union[DetBenchTrain, DetBenchPredict],
    dataset: Dataset,
    detection_threshold: float = 0.5,
    device: Optional[torch.device] = None,
) -> List[Prediction]:

    batch, records = build_infer_batch(dataset)
    return _predict_batch(
        model=model,
        batch=batch,
        records=records,
        detection_threshold=detection_threshold,
        device=device,
    )


def predict_dl(
    model: nn.Module, infer_dl: DataLoader, show_pbar: bool = True, **predict_kwargs
):
    return _predict_dl(
        predict_fn=_predict_batch,
        model=model,
        infer_dl=infer_dl,
        show_pbar=show_pbar,
        **predict_kwargs,
    )


def convert_raw_predictions(
    raw_preds: torch.Tensor, records: Sequence[BaseRecord], detection_threshold: float
) -> List[Prediction]:
    dets = raw_preds.detach().cpu().numpy()
    preds = []
    for det, record in zip(dets, records):
        if detection_threshold > 0:
            scores = det[:, 4]
            keep = scores > detection_threshold
            det = det[keep]

        pred = BaseRecord(
            (
                ScoresRecordComponent(),
                ImageRecordComponent(),
                InstancesLabelsRecordComponent(),
                BBoxesRecordComponent(),
            )
        )

        pred.detection.set_class_map(record.detection.class_map)
        pred.detection.set_labels_by_id(det[:, 5].astype(int))
        pred.detection.set_bboxes([BBox.from_xyxy(*xyxy) for xyxy in det[:, :4]])
        pred.detection.set_scores(det[:, 4])

        preds.append(Prediction(pred=pred, ground_truth=record))

    return preds



__all__ = [
    "EfficientDetBackboneConfig",
]

from icevision.imports import *
from icevision.backbones import BackboneConfig


class EfficientDetBackboneConfig(BackboneConfig):
    def __init__(self, model_name: str):
        self.model_name = model_name
        self.pretrained: bool

    def __call__(self, pretrained: bool = True) -> "EfficientDetBackboneConfig":
        """Completes the configuration of the backbone

        # Arguments
            pretrained: If True, use a pretrained backbone (on COCO).
        """
        self.pretrained = pretrained
        return self



__all__ = ["base_show_results"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.visualize import *
from icevision.data import *


def base_show_results(
    predict_fn: callable,
    model: nn.Module,
    dataset: Dataset,
    num_samples: int = 6,
    ncols: int = 3,
    denormalize_fn: Optional[callable] = denormalize_imagenet,
    show: bool = True,
    **predict_kwargs,
) -> None:
    records = random.choices(dataset, k=num_samples)
    preds = predict_fn(model, records, **predict_kwargs)

    show_preds(
        preds,
        denormalize_fn=denormalize_fn,
        ncols=ncols,
        show=show,
    )



__all__ = ["download_mmdet_configs"]

from icevision.imports import *
from icevision.utils import *

VERSION = "v2.10.0"
BASE_URL = "https://codeload.github.com/lgvaz/mmdetection_configs/zip/refs/tags"


def download_mmdet_configs() -> Path:
    save_dir = get_root_dir() / f"mmdetection_configs"
    save_dir.mkdir(parents=True, exist_ok=True)

    download_path = save_dir / f"{VERSION}.zip"
    if not download_path.exists():
        logger.info("Downloading mmdet configs")

        download_and_extract(f"{BASE_URL}/{VERSION}", download_path)

    return save_dir / f"mmdetection_configs-{VERSION[1:]}/configs"



__all__ = ["BackboneConfig"]

from icevision.imports import *


class BackboneConfig(ABC):
    @abstractmethod
    def __call__(self, *args: Any, **kwargs: Any) -> "BackboneConfig":
        """Completes configuration for this backbone.

        Called by the end user. All heavy lifting (such as creating a NN backbone)
        should be done here.
        """



__all__ = ["Component", "Composite", "TaskComponent", "TaskComposite"]

from icevision.imports import *
from icevision.core import tasks


class Component:
    order = 0.5

    def __init__(self):
        self.composite = None

    def set_composite(self, composite):
        self.composite = composite


class TaskComponent(Component):
    def __init__(self, task=tasks.common):
        self.task = task


class TaskComposite:
    base_components = set()

    def __init__(self, components: Sequence[TaskComponent]):
        components = set(components)
        components.update(set(comp() for comp in self.base_components))
        self.components = components
        self.set_task_components(self.components)

    def __getattr__(self, name):
        if name == "task_composites":
            raise AttributeError(name)

        # TODO: Possible bug if no task with _default is passed
        try:
            return getattr(self.task_composites[tasks.common.name], name)
        except AttributeError:
            pass

        try:
            return self.task_composites[name]
        except KeyError:
            pass

        raise AttributeError(f"{self.__class__.__name__} has no attribute {name}")

    def add_component(self, component: TaskComponent):
        self.components.add(component)
        self.set_task_components(self.components)

    def set_task_components(self, components: Sequence[TaskComponent]):
        task_components = defaultdict(list)
        # example: task_components['detect'] = (LabelsComponent, BBoxesComponent, ...)
        for component in components:
            task_components[component.task].append(component)

        self.task_composites = OrderedDict()
        for task, components in sorted(
            task_components.items(), key=lambda o: o[0].order
        ):
            self.task_composites[task.name] = composite = Composite()
            composite.add_components(components)
            if task != tasks.common:
                composite.set_parent(self)

    # TODO: rename reduce_on_all_tasks_components
    def reduce_on_components(
        self,
        fn_name: str,
        reduction: Optional[str] = None,
        **fn_kwargs,
    ) -> Dict[str, Any]:
        results = {}
        for task, composite in self.task_composites.items():
            result = composite.reduce_on_components(fn_name, reduction, **fn_kwargs)
            results[task] = result

        return results

    def reduce_on_task_components(
        self,
        fn_name: str,
        task_name: str,
        reduction: Optional[str] = None,
        **fn_kwargs,
    ) -> Any:
        composite = self.task_composites[task_name]
        return composite.reduce_on_components(fn_name, reduction, **fn_kwargs)


class Composite:
    base_components = set()

    def __init__(self, components: Optional[Sequence[Component]] = None):
        self._parent = None

        components = set(components) if components is not None else set()
        components.update(set(comp() for comp in self.base_components))
        self.set_components(components)

    def __getattr__(self, name):
        # avoid recursion https://nedbatchelder.com/blog/201010/surprising_getattr_recursion.html
        if name in ["components", "_parent"]:
            raise AttributeError(name)
        # delegates attributes to components
        for component in self.components:
            try:
                return getattr(component, name)
            except AttributeError:
                pass
        # delegates attributes to parent
        try:
            return getattr(self._parent, name)
        except AttributeError:
            pass

        raise AttributeError(f"{self.__class__.__name__} has no attribute {name}")

    def reduce_on_components(
        self, fn_name: str, reduction: Optional[str] = None, **fn_kwargs
    ) -> Any:
        results = []
        for component in self.components:
            results.append(getattr(component, fn_name)(**fn_kwargs))

        if reduction is not None and len(results) > 0:
            out = results.pop(0)
            for r in results:
                getattr(out, reduction)(r)
        else:
            out = results

        return out

    def get_component_by_type(self, component_type) -> Union[Component, None]:
        for component in self.components:
            if isinstance(component, component_type):
                return component

    def add_component(self, component):
        self.add_components([component])

    def add_components(self, components):
        self.set_components(set(components).union(set(self.components)))

    def set_components(self, components):
        self.components = sorted(components, key=lambda o: o.order)
        self.components_cls = [comp.__class__ for comp in self.components]

        for component in self.components:
            component.set_composite(self)

    def set_parent(self, parent):
        self._parent = parent



__all__ = ["TorchvisionBackboneConfig"]

from icevision.imports import *
from icevision.backbones import BackboneConfig


class TorchvisionBackboneConfig(BackboneConfig):
    def __init__(self, backbone_fn, **backbone_fn_kwargs):
        self.backbone: nn.Module

        self.backbone_fn = backbone_fn
        self.backbone_fn_kwargs = backbone_fn_kwargs

    def __call__(self, pretrained: bool = True, **kwargs):
        """Completes the configuration of the backbone

        # Arguments
            pretrained: If True, use a pretrained backbone (on COCO).
            By default it is set to True: this is generally used when training a new model (transfer learning).
            `pretrained = False`  is used during inference (prediction) for cases where the users have their own pretrained weights.
        """
        self.pretrained = pretrained

        # kwargs passed to call overwrite kwargs from init
        kwargs = {**self.backbone_fn_kwargs, **kwargs}
        self.backbone = self.backbone_fn(pretrained=pretrained, **kwargs)

        return self



__all__ = [
    "tf_lite0",
    "d0",
    "d1",
    "d2",
    "d3",
    "d4",
    "d5",
    "d6",
    "d7",
    "d7x",
]

from icevision.models.ross.efficientdet.utils import *


tf_lite0 = EfficientDetBackboneConfig(model_name="tf_efficientdet_lite0")

d0 = EfficientDetBackboneConfig(model_name="efficientdet_d0")

d1 = EfficientDetBackboneConfig(model_name="efficientdet_d1")

d2 = EfficientDetBackboneConfig(model_name="efficientdet_d2")

d3 = EfficientDetBackboneConfig(model_name="efficientdet_d3")

d4 = EfficientDetBackboneConfig(model_name="efficientdet_d4")

d5 = EfficientDetBackboneConfig(model_name="efficientdet_d5")

d6 = EfficientDetBackboneConfig(model_name="efficientdet_d6")

d7 = EfficientDetBackboneConfig(model_name="efficientdet_d7")

d7x = EfficientDetBackboneConfig(model_name="efficientdet_d7x")



__all__ = [
    "MMDetBackboneConfig",
    "mmdet_configs_path",
    "param_groups",
    "MMDetBackboneConfig",
    "create_model_config",
]

from icevision.imports import *
from icevision.utils import *
from icevision.backbones import BackboneConfig
from icevision.models.mmdet.download_configs import download_mmdet_configs
from mmdet.models.detectors import *
from mmcv import Config


mmdet_configs_path = download_mmdet_configs()


class MMDetBackboneConfig(BackboneConfig):
    def __init__(self, model_name, config_path, weights_url):
        self.model_name = model_name
        self.config_path = config_path
        self.weights_url = weights_url
        self.pretrained: bool

    def __call__(self, pretrained: bool = True) -> "MMDetBackboneConfig":
        self.pretrained = pretrained
        return self


def param_groups(model):
    body = model.backbone

    layers = []
    layers += [nn.Sequential(body.conv1, body.bn1)]
    layers += [getattr(body, l) for l in body.res_layers]
    layers += [model.neck]

    if isinstance(model, SingleStageDetector):
        layers += [model.bbox_head]
    elif isinstance(model, TwoStageDetector):
        layers += [nn.Sequential(model.rpn_head, model.roi_head)]
    else:
        raise RuntimeError(
            "{model} must inherit either from SingleStageDetector or TwoStageDetector class"
        )

    _param_groups = [list(layer.parameters()) for layer in layers]
    check_all_model_params_in_groups2(model, _param_groups)
    return _param_groups


def create_model_config(
    backbone: MMDetBackboneConfig,
    pretrained: bool = True,
    checkpoints_path: Optional[Union[str, Path]] = "checkpoints",
    force_download=False,
):

    model_name = backbone.model_name
    config_path = backbone.config_path
    weights_url = backbone.weights_url

    # download weights
    if pretrained and weights_url:
        save_dir = Path(checkpoints_path) / model_name
        save_dir.mkdir(exist_ok=True, parents=True)

        fname = Path(weights_url).name
        weights_path = save_dir / fname

        if not weights_path.exists() or force_download:
            download_url(url=weights_url, save_path=str(weights_path))

    cfg = Config.fromfile(config_path)

    return cfg, weights_path



__all__ = [
    "resnet50_caffe_fpn_1x",
    "resnet50_fpn_1x",
    "resnet50_fpn_2x",
    "resnet101_caffe_fpn_1x",
    "resnet101_fpn_1x",
    "resnet101_fpn_2x",
    "resnext101_32x4d_fpn_1x",
    "resnext101_32x4d_fpn_2x",
    "resnext101_64x4d_fpn_1x",
    "resnext101_64x4d_fpn_2x",
]

from icevision.imports import *
from icevision.models.mmdet.utils import *


class MMDetRetinanetBackboneConfig(MMDetBackboneConfig):
    def __init__(self, **kwargs):
        super().__init__(model_name="retinanet", **kwargs)


base_config_path = mmdet_configs_path / "retinanet"
base_weights_url = "http://download.openmmlab.com/mmdetection/v2.0/retinanet"

resnet50_caffe_fpn_1x = MMDetRetinanetBackboneConfig(
    config_path=base_config_path / "retinanet_r50_caffe_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/retinanet_r50_caffe_fpn_1x_coco/retinanet_r50_caffe_fpn_1x_coco_20200531-f11027c5.pth",
)

resnet50_fpn_1x = MMDetRetinanetBackboneConfig(
    config_path=base_config_path / "retinanet_r50_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/retinanet_r50_fpn_1x_coco/retinanet_r50_fpn_1x_coco_20200130-c2398f9e.pth",
)

resnet50_fpn_2x = MMDetRetinanetBackboneConfig(
    config_path=base_config_path / "retinanet_r50_fpn_2x_coco.py",
    weights_url=f"{base_weights_url}/retinanet_r50_fpn_1x_coco/retinanet_r50_fpn_1x_coco_20200130-c2398f9e.pth",
)

resnet101_caffe_fpn_1x = MMDetRetinanetBackboneConfig(
    config_path=base_config_path / "retinanet_r101_caffe_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/retinanet_r101_caffe_fpn_1x_coco/retinanet_r101_caffe_fpn_1x_coco_20200531-b428fa0f.pth",
)

resnet101_fpn_1x = MMDetRetinanetBackboneConfig(
    config_path=base_config_path / "retinanet_r101_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/retinanet_r101_fpn_1x_coco/retinanet_r101_fpn_1x_coco_20200130-7a93545f.pth",
)

resnet101_fpn_2x = MMDetRetinanetBackboneConfig(
    config_path=base_config_path / "retinanet_r101_fpn_2x_coco.py",
    weights_url=f"{base_weights_url}/retinanet_r101_fpn_2x_coco/retinanet_r101_fpn_2x_coco_20200131-5560aee8.pth",
)

resnext101_32x4d_fpn_1x = MMDetRetinanetBackboneConfig(
    config_path=base_config_path / "retinanet_x101_32x4d_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/retinanet_x101_32x4d_fpn_1x_coco/retinanet_x101_32x4d_fpn_1x_coco_20200130-5c8b7ec4.pth",
)

resnext101_32x4d_fpn_2x = MMDetRetinanetBackboneConfig(
    config_path=base_config_path / "retinanet_x101_32x4d_fpn_2x_coco.py",
    weights_url=f"{base_weights_url}/retinanet_x101_32x4d_fpn_2x_coco/retinanet_x101_32x4d_fpn_2x_coco_20200131-237fc5e1.pth",
)

resnext101_64x4d_fpn_1x = MMDetRetinanetBackboneConfig(
    config_path=base_config_path / "retinanet_x101_64x4d_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/retinanet_x101_64x4d_fpn_1x_coco/retinanet_x101_64x4d_fpn_1x_coco_20200130-366f5af1.pth",
)

resnext101_64x4d_fpn_2x = MMDetRetinanetBackboneConfig(
    config_path=base_config_path / "retinanet_x101_64x4d_fpn_2x_coco.py",
    weights_url=f"{base_weights_url}/retinanet_x101_64x4d_fpn_2x_coco/retinanet_x101_64x4d_fpn_2x_coco_20200131-bca068ab.pth",
)



__all__ = [
    "remove_internal_model_transforms",
    "patch_rcnn_param_groups",
    "patch_retinanet_param_groups",
]

from icevision.imports import *
from icevision.utils import *
from torchvision.models.detection.generalized_rcnn import GeneralizedRCNN


def _noop_normalize(image: Tensor) -> Tensor:
    return image


def _noop_resize(
    image: Tensor, target: Optional[Dict[str, Tensor]]
) -> Tuple[Tensor, Optional[Dict[str, Tensor]]]:
    return image, target


def remove_internal_model_transforms(model: GeneralizedRCNN):
    model.transform.normalize = _noop_normalize
    model.transform.resize = _noop_resize


def patch_param_groups(
    model: nn.Module,
    head_layers: List[nn.Module],
    backbone_param_groups: List[List[nn.Parameter]],
):
    def param_groups(model: nn.Module) -> List[List[nn.Parameter]]:
        head_param_groups = [list(layer.parameters()) for layer in head_layers]

        _param_groups = backbone_param_groups + head_param_groups
        check_all_model_params_in_groups2(model, _param_groups)

        return _param_groups

    model.param_groups = MethodType(param_groups, model)


def patch_rcnn_param_groups(model: nn.Module):
    return patch_param_groups(
        model=model,
        head_layers=[model.rpn, model.roi_heads],
        backbone_param_groups=model.backbone.param_groups(),
    )


def patch_retinanet_param_groups(model: nn.Module):
    return patch_param_groups(
        model=model,
        head_layers=[model.head],
        backbone_param_groups=model.backbone.param_groups(),
    )



__all__ = ["aggregate_records_objects"]

from icevision.imports import *
from .record import BaseRecord


def aggregate_records_objects(
    records: List[BaseRecord],
) -> Dict[str, List[Dict[str, Any]]]:
    objects = defaultdict(list)
    for record in records:
        for k, v in record.aggregate_objects().items():
            objects[k].extend(v)
    return dict(objects)



__all__ = ["Task", "common", "detection", "classification"]

from icevision.imports import *


@dataclass(eq=True, frozen=True)
class Task:
    name: str
    order: float = 0.5


common = Task("common", 0.3)
detection = Task("detection")
classification = Task("classification")



__all__ = [
    "show_sample",
    "show_record",
    "show_pred",
    "show_samples",
    "show_records",
    "show_preds",
]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.parsers import *
from icevision.data import *
from icevision.visualize.draw_data import *


def show_sample(
    sample,
    class_map: Optional[ClassMap] = None,
    denormalize_fn: Optional[callable] = denormalize_imagenet,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    display_keypoints: bool = True,
    show: bool = True,
    ax: plt.Axes = None,
    figsize: Tuple[int, int] = None,
    **draw_sample_kwargs,
) -> None:
    img = draw_sample(
        sample=sample,
        class_map=class_map,
        denormalize_fn=denormalize_fn,
        display_label=display_label,
        display_bbox=display_bbox,
        display_mask=display_mask,
        display_keypoints=display_keypoints,
        **draw_sample_kwargs,
    )
    show_img(img=img, ax=ax, show=show, figsize=figsize)


def show_record(
    record,
    class_map: Optional[ClassMap] = None,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    display_keypoints: bool = True,
    ax: plt.Axes = None,
    show: bool = False,
    figsize: Tuple[int, int] = None,
    **draw_sample_kwargs,
) -> None:
    img = draw_record(
        record=record,
        class_map=class_map,
        display_label=display_label,
        display_bbox=display_bbox,
        display_mask=display_mask,
        display_keypoints=display_keypoints,
        **draw_sample_kwargs,
    )
    show_img(img=img, ax=ax, show=show, figsize=figsize)


def show_pred(
    pred: Prediction,
    denormalize_fn: Optional[callable] = denormalize_imagenet,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    show: bool = True,
    ax: plt.Axes = None,
    figsize=None,
    annotation=None,
    **draw_sample_kwargs,
) -> None:
    actual = draw_sample(
        sample=pred.ground_truth,
        display_label=display_label,
        display_bbox=display_bbox,
        display_mask=display_mask,
        denormalize_fn=denormalize_fn,
    )

    prediction = draw_pred(
        pred=pred,
        denormalize_fn=denormalize_fn,
        display_label=display_label,
        display_bbox=display_bbox,
        display_mask=display_mask,
        **draw_sample_kwargs,
    )

    if ax is None:
        fig, ax = plt.subplots(1, 2, figsize=figsize)

    ax[0].imshow(actual, cmap=None)
    ax[0].set_title("Ground truth")
    ax[1].imshow(prediction, cmap=None)
    ax[1].set_title("Prediction")

    if annotation is None:
        ax[0].set_axis_off()
        ax[1].set_axis_off()
    else:
        ax[0].get_xaxis().set_ticks([])
        ax[0].set_frame_on(False)
        ax[0].get_yaxis().set_visible(False)
        ax[0].set_xlabel(annotation[0], ma="left")

        ax[1].get_xaxis().set_ticks([])
        ax[1].set_frame_on(False)
        ax[1].get_yaxis().set_visible(False)
        ax[1].set_xlabel(annotation[1], ma="left")

    if show:
        plt.show()


def show_records(
    records: Sequence[RecordType],
    class_map: Optional[ClassMap] = None,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    ncols: int = 1,
    figsize=None,
    show: bool = False,
    **draw_sample_kwargs,
) -> None:
    partials = [
        partial(
            show_record,
            record=record,
            display_label=display_label,
            display_bbox=display_bbox,
            display_mask=display_mask,
            class_map=class_map,
            show=False,
            **draw_sample_kwargs,
        )
        for record in records
    ]
    plot_grid(partials, ncols=ncols, figsize=figsize, show=show)


def show_samples(
    samples: Sequence[dict],
    class_map: Optional[ClassMap] = None,
    denormalize_fn: Optional[callable] = denormalize_imagenet,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    ncols: int = 1,
    figsize=None,
    show=False,
    **draw_sample_kwargs,
) -> None:
    partials = [
        partial(
            show_sample,
            sample=sample,
            denormalize_fn=denormalize_fn,
            display_label=display_label,
            display_bbox=display_bbox,
            display_mask=display_mask,
            class_map=class_map,
            show=False,
            **draw_sample_kwargs,
        )
        for sample in samples
    ]
    plot_grid(partials, ncols=ncols, figsize=figsize, show=show)


def show_preds(
    preds: Sequence[Prediction],
    denormalize_fn: Optional[callable] = denormalize_imagenet,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    figsize=None,
    show=False,
    annotations=None,
    ncols=None,  # TODO: Not used,
    **draw_sample_kwargs,
) -> None:
    annotations = annotations or [None] * len(preds)
    partials = [
        partial(
            show_pred,
            pred=pred,
            denormalize_fn=denormalize_fn,
            display_label=display_label,
            display_bbox=display_bbox,
            display_mask=display_mask,
            show=False,
            annotation=annotation,
            **draw_sample_kwargs,
        )
        for pred, annotation in zip(preds, annotations)
    ]

    plot_grid(
        partials,
        ncols=2,
        figsize=figsize or (6, 6 * len(preds) / 2 / 0.75),
        show=show,
        axs_per_iter=2,
    )



import sys, os, re, shutil, typing, itertools, operator, math, warnings, json, random
import functools, io, cv2, mimetypes, torch, torchvision, dataclasses, zipfile, pickle
from copy import copy, deepcopy

from pdb import set_trace
from dataclasses import dataclass
from fastcore.foundation import *
from fastcore.basics import *

import numpy as np
import torch.nn as nn
import torch.optim.lr_scheduler as lr_scheduler
import matplotlib.pyplot as plt

from types import (
    BuiltinFunctionType,
    BuiltinMethodType,
    MethodType,
    FunctionType,
    SimpleNamespace,
)

from abc import (
    ABC,
    abstractmethod,
    abstractproperty,
    abstractclassmethod,
    abstractstaticmethod,
)
from pathlib import Path
from collections import defaultdict, OrderedDict, namedtuple
from enum import Enum
from tqdm.auto import tqdm
from pycocotools import mask as mask_utils
from contextlib import contextmanager
from typing import *
from operator import itemgetter, attrgetter
from functools import partial, reduce

from torch import tensor, Tensor
from torch.utils.data import DataLoader
from torch.optim import SGD, Adam, AdamW, Adagrad, Adadelta, RMSprop
from torch.optim.lr_scheduler import (
    LambdaLR,
    StepLR,
    MultiStepLR,
    MultiplicativeLR,
    OneCycleLR,
    CosineAnnealingLR,
    CosineAnnealingWarmRestarts,
)

from torchvision.transforms.functional import to_tensor as im2tensor

from loguru import logger

# Soft imports
from icevision.soft_dependencies import SoftDependencies

if SoftDependencies.fastai:
    import fastai.vision.all as fastai

if SoftDependencies.pytorch_lightning:
    import pytorch_lightning as pl

if SoftDependencies.wandb:
    import wandb


# TODO: Stop importing partial from fastcore and move this to utils
class partial:
    """Wraps functools.partial, same functionality.

    Modifies the original partial `__repr__` and `__str__` in other to fix #270
    """

    def __init__(self, func, *args, **kwargs):
        self._partial = functools.partial(func, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self._partial(*args, **kwargs)

    def __str__(self):
        name = self._partial.func.__name__
        partial_str = str(self._partial)
        return re.sub(r"<.+>", name, partial_str)

    def __repr__(self):
        return str(self)



__all__ = [
    "Mask",
    "MaskArray",
    "MaskFile",
    "VocMaskFile",
    "RLE",
    "Polygon",
    "EncodedRLEs",
]

from icevision.imports import *
from icevision.utils import *
from PIL import Image


class Mask(ABC):
    @abstractmethod
    def to_mask(self, h, w) -> "MaskArray":
        pass

    @abstractmethod
    def to_erles(self, h, w) -> "EncodedRLEs":
        pass


class EncodedRLEs(Mask):
    def __init__(self, erles: List[dict] = None):
        self.erles = erles or []

    def __repr__(self):
        return f"<{self.__class__.__name__} with {len(self)} objects>"

    def __len__(self):
        return len(self.erles)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.erles == other.erles
        return False

    def append(self, v: "EncodedRLEs"):
        self.erles.extend(v.erles)

    def extend(self, v: List["EncodedRLEs"]):
        for o in v:
            self.append(o)

    def pop(self, i: int):
        self.erles.pop(i)

    def to_mask(self, h, w) -> "MaskArray":
        mask = mask_utils.decode(self.erles)
        mask = mask.transpose(2, 0, 1)  # channels first
        return MaskArray(mask)

    def to_erles(self, h, w) -> "EncodedRLEs":
        return self


# TODO: Assert shape? (bs, height, width)
class MaskArray(Mask):
    """Binary numpy array representation of a mask.

    # Arguments
        data: Mask array, with the dimensions: (num_instances, height, width)
    """

    def __init__(self, data: np.uint8):
        self.data = data.astype(np.uint8)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, i):
        return type(self)(self.data[i])

    def to_tensor(self):
        return tensor(self.data, dtype=torch.uint8)

    def to_mask(self, h, w):
        return self

    def to_erles(self, h, w) -> EncodedRLEs:
        return EncodedRLEs(
            mask_utils.encode(np.asfortranarray(self.data.transpose(1, 2, 0)))
        )

    def to_coco_rle(self, h, w) -> List[dict]:
        """From https://stackoverflow.com/a/49547872/6772672"""
        assert self.data.shape[1:] == (h, w)
        rles = []
        for mask in self.data:
            counts = []
            flat = itertools.groupby(mask.ravel(order="F"))
            for i, (value, elements) in enumerate(flat):
                if i == 0 and value == 1:
                    counts.append(0)
                counts.append(len(list(elements)))
            rles.append({"counts": counts, "size": (h, w)})
        return rles

    @property
    def shape(self):
        return self.data.shape

    @classmethod
    def from_masks(cls, masks: Union[EncodedRLEs, Sequence[Mask]], h: int, w: int):
        # HACK: check for backwards compatibility
        if isinstance(masks, EncodedRLEs):
            return masks.to_mask(h, w)
        else:
            masks_arrays = [o.to_mask(h=h, w=w).data for o in masks]
            return cls(np.concatenate(masks_arrays))


class MaskFile(Mask):
    """Holds the path to mask image file.

    # Arguments
        filepath: Path to the mask image file.
    """

    def __init__(self, filepath: Union[str, Path]):
        self.filepath = Path(filepath)

    def to_mask(self, h, w):
        mask = open_img(self.filepath, gray=True)
        obj_ids = np.unique(mask)[1:]
        masks = mask == obj_ids[:, None, None]
        return MaskArray(masks)

    def to_coco_rle(self, h, w) -> List[dict]:
        return self.to_mask(h=h, w=w).to_coco_rle(h=h, w=w)

    def to_erles(self, h, w) -> EncodedRLEs:
        return self.to_mask(h, w).to_erles(h, w)


class VocMaskFile(MaskFile):
    """Extension of `MaskFile` for VOC masks.
    Removes the color pallete and optionally drops void pixels.

    # Arguments
        drop_void (bool): drops the void pixels, which should have the value 255.
        filepath: Path to the mask image file.
    """

    def __init__(self, filepath: Union[str, Path], drop_void: bool = True):
        super().__init__(filepath=filepath)
        self.drop_void = drop_void

    def to_mask(self, h, w) -> MaskArray:
        mask_arr = np.array(Image.open(self.filepath))
        obj_ids = np.unique(mask_arr)[1:]
        masks = mask_arr == obj_ids[:, None, None]

        if self.drop_void:
            masks = masks[:-1, ...]

        return MaskArray(masks)


class RLE(Mask):
    """Run length encoding of a mask.

    Don't instantiate this class directly, instead use the classmethods
    `from_coco` and `from_kaggle`.
    """

    def __init__(self, counts: List[int]):
        self.counts = counts

    def to_mask(self, h, w) -> "MaskArray":
        return self.to_erles(h=h, w=w).to_mask(h=h, w=w)
        # Convert kaggle counts to mask
        # "From https://www.kaggle.com/julienbeaulieu/imaterialist-detectron2"
        # mask = np.full(h * w, 0, dtype=np.uint8)
        # for start, ones in zip(self.counts[::2], self.counts[1::2]):
        #     # counting starts on one
        #     start -= 1
        #     if ones:
        #         mask[start : start + ones] = 1
        # mask = mask.reshape((h, w), order="F")
        # return MaskArray(mask)

    def to_coco(self) -> List[int]:
        return self.counts

    def to_erles(self, h, w) -> EncodedRLEs:
        return EncodedRLEs(
            mask_utils.frPyObjects([{"counts": self.to_coco(), "size": [h, w]}], h, w)
        )

    @classmethod
    def from_string(cls, s, sep=" "):
        return cls(lmap(int, s.split(sep)))

    @classmethod
    def from_kaggle(cls, counts: Sequence[int]):
        """Described [here](https://www.kaggle.com/c/imaterialist-fashion-2020-fgvc7/overview/evaluation)"""
        if len(counts) % 2 != 0:
            raise ValueError("Counts must be divisible by 2")

        current = 1
        coco_counts = []
        for start, count in zip(counts[::2], counts[1::2]):
            coco_counts.append(start - current)  # zeros
            coco_counts.append(count)  # ones
            current = start + count

        # remove trailing zero
        if coco_counts[-1] == 0:
            coco_counts.pop(-1)

        return cls.from_coco(coco_counts)

    @classmethod
    def from_coco(cls, counts: Sequence[int]):
        """Described [here](https://stackoverflow.com/a/49547872/6772672)"""
        return cls(counts)
        # Convert from kaggle to coco
        # when counts is odd, round it with 0 ones at the end
        # if len(counts) % 2 != 0:
        #     counts = counts + [0]
        #
        # kaggle_counts, total = [], 0
        # for zeros, ones in zip(counts[::2], counts[1::2]):
        #     start = zeros + total + 1
        #     kaggle_counts.extend([start, ones])
        #     total += zeros + ones
        # return cls.from_kaggle(kaggle_counts)


class Polygon(Mask):
    """Polygon representation of a mask

    # Arguments
        points: The vertices of the polygon in the COCO standard format.
    """

    def __init__(self, points: List[List[int]]):
        self.points = points

    def to_mask(self, h, w):
        return self.to_erles(h=h, w=w).to_mask(h=h, w=w)

    def to_erles(self, h, w) -> EncodedRLEs:
        erles = mask_utils.frPyObjects(self.points, h, w)
        erle = mask_utils.merge(erles)  # make unconnected polygons a single mask
        return EncodedRLEs([erle])



__all__ = ["mobilenet"]

from icevision.backbones import mobilenet as mobilenet_fn
from icevision.models.torchvision.backbones.backbone_config import (
    TorchvisionBackboneConfig,
)

mobilenet = TorchvisionBackboneConfig(mobilenet_fn)



from icevision.models.torchvision.backbones.resnet_configs import *
from icevision.models.torchvision.backbones.resnet_fpn_configs import *
from icevision.models.torchvision.backbones.mobilenet_configs import *



__all__ = ["Transform"]

from icevision.imports import *
from icevision.core import *


class Transform(ABC):
    def __call__(self, record: BaseRecord):
        # TODO: this assumes record is already loaded and copied
        # which is generally true
        return self.apply(record)

    @abstractmethod
    def apply(self, record: BaseRecord) -> BaseRecord:
        """Apply the transform
        Returns:
              dict: Modified values, the keys of the dictionary should have the same
              names as the keys received by this function
        """



__all__ = ["model"]

from icevision.imports import *
from mmcv import Config
from mmdet.models import build_detector
from mmcv.runner import load_checkpoint
from icevision.models.mmdet.utils import *
from icevision.models.mmdet.common.utils import *


def model(
    backbone: MMDetBackboneConfig,
    num_classes: int,
    checkpoints_path: Optional[Union[str, Path]] = "checkpoints",
    force_download=False,
) -> nn.Module:

    return build_model(
        model_type="two_stage_detector_mask",
        backbone=backbone,
        num_classes=num_classes,
        pretrained=backbone.pretrained,
        checkpoints_path=checkpoints_path,
        force_download=force_download,
    )



from icevision.models.mmdet.models.sparse_rcnn import backbones
from icevision.models.mmdet.common.bbox.two_stage import *
from icevision.models.interpretation import Interpretation, _move_to_device
from icevision.models.mmdet.common.interpretation_utils import (
    sum_losses_mmdet,
    loop_mmdet,
)

_LOSSES_DICT = {
    "loss_cls": [],
    "loss_bbox": [],
    "loss_total": [],
}

interp = Interpretation(
    losses_dict=_LOSSES_DICT,
    valid_dl=valid_dl,
    infer_dl=infer_dl,
    predict_dl=predict_dl,
)

interp._loop = loop_mmdet



__all__ = [
    "RecordComponent",
    "ClassMapRecordComponent",
    "RecordIDRecordComponent",
    "ImageRecordComponent",
    "FilepathRecordComponent",
    "SizeRecordComponent",
    "BaseLabelsRecordComponent",
    "InstancesLabelsRecordComponent",
    "ClassificationLabelsRecordComponent",
    "BBoxesRecordComponent",
    "MasksRecordComponent",
    "AreasRecordComponent",
    "IsCrowdsRecordComponent",
    "KeyPointsRecordComponent",
    "ScoresRecordComponent",
    "LossesRecordComponent",
]

from icevision.imports import *
from icevision.utils import *
from icevision.core.components import *
from icevision.core.bbox import *
from icevision.core.mask import *
from icevision.core.exceptions import *
from icevision.core.keypoints import *
from icevision.core.class_map import *
from icevision.core import tasks


class RecordComponent(TaskComponent):
    # TODO: as_dict is only necessary because of backwards compatibility
    @property
    def record(self):
        return self.composite

    def as_dict(self) -> dict:
        return {}

    def _load(self) -> None:
        return

    def _unload(self) -> None:
        return

    def _num_annotations(self) -> Dict[str, int]:
        return {}

    def _autofix(self) -> Dict[str, bool]:
        return {}

    def _remove_annotation(self, i) -> None:
        return

    def _aggregate_objects(self) -> Dict[str, List[dict]]:
        return {}

    def _repr(self) -> List[str]:
        return []

    def builder_template(self) -> List[str]:
        return self._format_builder_template(self._builder_template())

    def _builder_template(self) -> List[str]:
        return []

    def _format_builder_template(self, lines):
        task = f".{self.task.name}." if self.task != tasks.common else "."
        return [line.format(task=task) for line in lines]

    def setup_transform(self, tfm) -> None:
        pass


class ClassMapRecordComponent(RecordComponent):
    def __init__(self, task):
        super().__init__(task=task)
        self.class_map = None

    def set_class_map(self, class_map: ClassMap):
        self.class_map = class_map

    def _repr(self) -> List[str]:
        return [f"Class Map: {self.class_map}"]

    def as_dict(self) -> dict:
        return {"class_map": self.class_map}

    def _builder_template(self) -> List[str]:
        return ["record{task}set_class_map(<ClassMap>)"]


class RecordIDRecordComponent(RecordComponent):
    def __init__(self, task=tasks.common):
        super().__init__(task=task)
        self.record_id = None

    def set_record_id(self, record_id: int):
        self.record_id = record_id

    def _repr(self) -> List[str]:
        return [f"Image ID: {self.record_id}"]

    def as_dict(self) -> dict:
        return {"record_id": self.record_id}


# TODO: we need a way to combine filepath and image mixin
# TODO: rename to ImageArrayRecordComponent
class ImageRecordComponent(RecordComponent):
    def __init__(self, task=tasks.common):
        super().__init__(task=task)
        self.img = None

    def set_img(self, img: np.ndarray):
        self.img = img
        height, width, _ = self.img.shape
        # this should set on SizeRecordComponent
        self.composite.set_img_size(ImgSize(width=width, height=height), original=True)

    def _repr(self) -> List[str]:
        if self.img is not None:
            ndims = len(self.img.shape)
            if ndims == 3:  # RGB, RGBA
                height, width, channels = self.img.shape
            elif ndims == 2:  # Grayscale
                height, width, channels = [*self.img.shape, 1]
            else:
                raise ValueError(
                    f"Expected image to have 2 or 3 dimensions, got {ndims} instead"
                )
            return [f"Image: {width}x{height}x{channels} <np.ndarray> Image"]
        else:
            return [f"Image: {self.img}"]

    def _unload(self):
        self.img = None

    def as_dict(self) -> dict:
        return {"img": self.img}

    def setup_transform(self, tfm) -> None:
        tfm.setup_img(self)


class FilepathRecordComponent(ImageRecordComponent):
    def __init__(self, task=tasks.common):
        super().__init__(task=task)
        self.filepath = None

    def set_filepath(self, filepath: Union[str, Path]):
        self.filepath = Path(filepath)

    def _load(self):
        img = open_img(self.filepath)
        self.set_img(img)

    def _autofix(self) -> Dict[str, bool]:
        exists = self.filepath.exists()
        if not exists:
            raise AutofixAbort(f"File '{self.filepath}' does not exist")

        return super()._autofix()

    def _repr(self) -> List[str]:
        return [f"Filepath: {self.filepath}", *super()._repr()]

    def as_dict(self) -> dict:
        return {"filepath": self.filepath, **super().as_dict()}

    def _builder_template(self) -> List[str]:
        return ["record{task}set_filepath(<Union[str, Path]>)"]


class SizeRecordComponent(RecordComponent):
    def __init__(self, task=tasks.common):
        super().__init__(task=task)
        self.img_size = None

    def set_image_size(self, width: int, height: int):
        # TODO: use ImgSize
        self.img_size = ImgSize(width=width, height=height)
        self.width, self.height = width, height

    def set_img_size(self, size: ImgSize, original: bool = False):
        self.img_size = size
        self.width, self.height = size

        if original:
            self.original_img_size = size

    def setup_transform(self, tfm) -> None:
        tfm.setup_size(self)

    def _repr(self) -> List[str]:
        return [
            f"Image size {self.img_size}",
        ]

    def as_dict(self) -> dict:
        return {"width": self.width, "height": self.height}

    def _aggregate_objects(self) -> Dict[str, List[dict]]:
        info = [{"img_width": self.width, "img_height": self.height}]
        return {"img_size": info}

    def _builder_template(self) -> List[str]:
        return ["record{task}set_img_size(<ImgSize>)"]


### Annotation parsers ###
class BaseLabelsRecordComponent(ClassMapRecordComponent):
    def __init__(self, task=tasks.common):
        super().__init__(task=task)
        self.label_ids: List[int] = []
        self.labels: List[Hashable] = []

    # TODO: rename to labels_ids
    def set_labels_by_id(self, labels: Sequence[int]):
        self.label_ids = list(labels)
        # TODO, HACK: necessary because `Dataset.from_images` has no class_map
        if self.class_map is not None:
            self.labels = self._labels_ids_to_names(labels)

    def add_labels_by_id(self, labels: Sequence[int]):
        self.label_ids.extend(labels)
        if self.class_map is not None:
            self.labels.extend(self._labels_ids_to_names(labels))

    def set_labels(self, labels_names: Sequence[Hashable]):
        self.labels = list(labels_names)
        self.label_ids = self._labels_names_to_ids(labels_names)

    def add_labels(self, labels_names: Sequence[Hashable]):
        self.labels.extend(labels_names)
        self.label_ids.extend(self._labels_names_to_ids(labels_names))

    def is_valid(self) -> List[bool]:
        return [True for _ in self.label_ids]

    def _labels_ids_to_names(self, labels_ids):
        return [self.class_map.get_by_id(id) for id in labels_ids]

    def _labels_names_to_ids(self, labels_names):
        return [self.class_map.get_by_name(name) for name in labels_names]

    def _num_annotations(self) -> Dict[str, int]:
        return {"labels": len(self.label_ids)}

    def _autofix(self) -> Dict[str, bool]:
        return {"labels": [True] * len(self.label_ids)}

    def _remove_annotation(self, i):
        self.label_ids.pop(i)

    def _aggregate_objects(self) -> Dict[str, List[dict]]:
        return {**super()._aggregate_objects(), "labels": self.label_ids}

    def _repr(self) -> List[str]:
        return [*super()._repr(), f"Labels: {self.label_ids}"]

    def as_dict(self) -> dict:
        return {"labels": self.label_ids}

    def _builder_template(self) -> List[str]:
        return [
            *super()._builder_template(),
            "record{task}add_labels(<Sequence[Hashable]>)",
        ]


class InstancesLabelsRecordComponent(BaseLabelsRecordComponent):
    def __init__(self, task=tasks.detection):
        super().__init__(task=task)

    def setup_transform(self, tfm) -> None:
        tfm.setup_instances_labels(self)


class ClassificationLabelsRecordComponent(BaseLabelsRecordComponent):
    def __init__(self, task=tasks.classification):
        super().__init__(task=task)


class BBoxesRecordComponent(RecordComponent):
    def __init__(self, task=tasks.detection):
        super().__init__(task=task)
        self.bboxes: List[BBox] = []

    def set_bboxes(self, bboxes: Sequence[BBox]):
        self.bboxes = list(bboxes)

    def add_bboxes(self, bboxes: Sequence[BBox]):
        self.bboxes.extend(bboxes)

    def _autofix(self) -> Dict[str, bool]:
        success = []
        for bbox in self.bboxes:
            try:
                autofixed = bbox.autofix(
                    img_w=self.composite.width, img_h=self.composite.height
                )
                success.append(True)
            except InvalidDataError as e:
                logger.log("AUTOFIX-FAIL", "{}", str(e))
                success.append(False)

        return {"bboxes": success}

    def _num_annotations(self) -> Dict[str, int]:
        return {"bboxes": len(self.bboxes)}

    def _remove_annotation(self, i):
        self.bboxes.pop(i)

    def _aggregate_objects(self) -> Dict[str, List[dict]]:
        objects = []
        for bbox in self.bboxes:
            x, y, w, h = bbox.xywh
            objects.append(
                {
                    "bbox_x": x,
                    "bbox_y": y,
                    "bbox_width": w,
                    "bbox_height": h,
                    "bbox_sqrt_area": bbox.area ** 0.5,
                    "bbox_aspect_ratio": w / h,
                }
            )

        return {"bboxes": objects}

    def _repr(self) -> List[str]:
        return [f"BBoxes: {self.bboxes}"]

    def as_dict(self) -> dict:
        return {"bboxes": self.bboxes}

    def setup_transform(self, tfm) -> None:
        tfm.setup_bboxes(self)

    def _builder_template(self) -> List[str]:
        return ["record{task}add_bboxes(<Sequence[BBox]>)"]


class MasksRecordComponent(RecordComponent):
    def __init__(self, task=tasks.detection):
        super().__init__(task=task)
        self.masks = EncodedRLEs()

    def set_masks(self, masks: Sequence[Mask]):
        self.masks = masks

    def add_masks(self, masks: Sequence[Mask]):
        self.masks.extend(self._masks_to_erle(masks))

    def setup_transform(self, tfm) -> None:
        tfm.setup_masks(self)

    def _masks_to_erle(self, masks: Sequence[Mask]) -> List[Mask]:
        width, height = self.composite.img_size
        return [mask.to_erles(h=height, w=width) for mask in masks]

    def _load(self):
        self._encoded_masks = self.masks
        self.masks = MaskArray.from_masks(
            self.masks, self.composite.height, self.composite.width
        )

    def _unload(self):
        self.masks = self._encoded_masks

    def _num_annotations(self) -> Dict[str, int]:
        return {"masks": len(self.masks)}

    def _remove_annotation(self, i):
        self.masks.pop(i)

    def _repr(self) -> List[str]:
        return [f"Masks: {self.masks}"]

    def as_dict(self) -> dict:
        return {"masks": self.masks}


class AreasRecordComponent(RecordComponent):
    def __init__(self, task=tasks.detection):
        super().__init__(task=task)
        self.areas: List[float] = []

    def set_areas(self, areas: Sequence[float]):
        self.areas = list(areas)

    def add_areas(self, areas: Sequence[float]):
        self.areas.extend(areas)

    def setup_transform(self, tfm) -> None:
        tfm.setup_areas(self)

    def _num_annotations(self) -> Dict[str, int]:
        return {"areas": len(self.areas)}

    def _remove_annotation(self, i):
        self.areas.pop(i)

    def _repr(self) -> List[str]:
        return [f"Areas: {self.areas}"]

    def as_dict(self) -> dict:
        return {"areas": self.areas}


class IsCrowdsRecordComponent(RecordComponent):
    def __init__(self, task=tasks.detection):
        super().__init__(task=task)
        self.iscrowds: List[bool] = []

    def set_iscrowds(self, iscrowds: Sequence[bool]):
        self.iscrowds = list(iscrowds)

    def add_iscrowds(self, iscrowds: Sequence[bool]):
        self.iscrowds.extend(iscrowds)

    def setup_transform(self, tfm) -> None:
        tfm.setup_iscrowds(self)

    def _num_annotations(self) -> Dict[str, int]:
        return {"iscrowds": len(self.iscrowds)}

    def _remove_annotation(self, i):
        self.iscrowds.pop(i)

    def _aggregate_objects(self) -> Dict[str, List[dict]]:
        return {"iscrowds": self.iscrowds}

    def _repr(self) -> List[str]:
        return [f"Is Crowds: {self.iscrowds}"]

    def as_dict(self) -> dict:
        return {"iscrowds": self.iscrowds}


class KeyPointsRecordComponent(RecordComponent):
    def __init__(self, task=tasks.detection):
        super().__init__(task=task)
        self.keypoints: List[KeyPoints] = []

    def set_keypoints(self, keypoints: Sequence[KeyPoints]):
        self.keypoints = list(keypoints)

    def add_keypoints(self, keypoints: Sequence[KeyPoints]):
        self.keypoints.extend(keypoints)

    def setup_transform(self, tfm) -> None:
        tfm.setup_keypoints(self)

    def as_dict(self) -> dict:
        return {"keypoints": self.keypoints}

    def _aggregate_objects(self) -> Dict[str, List[dict]]:
        objects = [
            {"keypoint_x": kpt.x, "keypoint_y": kpt.y, "keypoint_visible": kpt.v}
            for kpt in self.keypoints
        ]
        return {"keypoints": objects}

    def _repr(self) -> List[str]:
        return {f"KeyPoints: {self.keypoints}"}


class ScoresRecordComponent(RecordComponent):
    def __init__(self, task=tasks.detection):
        super().__init__(task=task)
        self.scores = None

    def set_scores(self, scores: Sequence[float]):
        self.scores = scores

    def _repr(self) -> List[str]:
        return [f"Scores: {self.scores}"]

    def as_dict(self) -> dict:
        return {"scores": self.scores}


class LossesRecordComponent(RecordComponent):
    def __init__(self, task=tasks.common):
        super().__init__(task=task)
        self.losses = None

    def set_losses(self, losses: Dict):
        self.losses = losses



__all__ = [
    "resnet50_fpn_1x",
    "resnet50_fpn_mstrain_480_800_3x",
    "resnet50_fpn_300_proposals_crop_mstrain_480_800_3x",
    "resnet101_fpn_mstrain_480_800_3x_coco",
    "resnet101_fpn_300_proposals_crop_mstrain_480_800_3x",
]

from icevision.imports import *
from icevision.models.mmdet.utils import *


class MMDetSparseRCNNBackboneConfig(MMDetBackboneConfig):
    def __init__(self, **kwargs):
        super().__init__(model_name="sparse_rcnn", **kwargs)


base_config_path = mmdet_configs_path / "sparse_rcnn"
base_weights_url = "https://download.openmmlab.com/mmdetection/v2.0/sparse_rcnn"

resnet50_fpn_1x = MMDetSparseRCNNBackboneConfig(
    config_path=base_config_path / "sparse_rcnn_r50_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/sparse_rcnn_r50_fpn_1x_coco/sparse_rcnn_r50_fpn_1x_coco_20201222_214453-dc79b137.pth",
)

resnet50_fpn_mstrain_480_800_3x = MMDetSparseRCNNBackboneConfig(
    config_path=base_config_path / "sparse_rcnn_r50_fpn_mstrain_480-800_3x_coco.py",
    weights_url=f"{base_weights_url}/sparse_rcnn_r50_fpn_mstrain_480-800_3x_coco/sparse_rcnn_r50_fpn_mstrain_480-800_3x_coco_20201218_154234-7bc5c054.pth",
)

resnet50_fpn_300_proposals_crop_mstrain_480_800_3x = MMDetSparseRCNNBackboneConfig(
    config_path=base_config_path
    / "sparse_rcnn_r50_fpn_300_proposals_crop_mstrain_480-800_3x_coco.py",
    weights_url=f"{base_weights_url}/sparse_rcnn_r50_fpn_300_proposals_crop_mstrain_480-800_3x_coco/sparse_rcnn_r50_fpn_300_proposals_crop_mstrain_480-800_3x_coco_20201223_024605-9fe92701.pth",
)

resnet101_fpn_mstrain_480_800_3x_coco = MMDetSparseRCNNBackboneConfig(
    config_path=base_config_path / "sparse_rcnn_r101_fpn_mstrain_480-800_3x_coco.py",
    weights_url=f"{base_weights_url}/sparse_rcnn_r101_fpn_mstrain_480-800_3x_coco/sparse_rcnn_r101_fpn_mstrain_480-800_3x_coco_20201223_121552-6c46c9d6.pth",
)

resnet101_fpn_300_proposals_crop_mstrain_480_800_3x = MMDetSparseRCNNBackboneConfig(
    config_path=base_config_path
    / "sparse_rcnn_r101_fpn_300_proposals_crop_mstrain_480-800_3x_coco.py",
    weights_url=f"{base_weights_url}/sparse_rcnn_r101_fpn_300_proposals_crop_mstrain_480-800_3x_coco/sparse_rcnn_r101_fpn_300_proposals_crop_mstrain_480-800_3x_coco_20201223_023452-c23c3564.pth",
)



__all__ = [
    "param_groups",
    "patch_param_groups",
]

from icevision.imports import *
from icevision.utils import *


def param_groups(model: nn.Module) -> List[nn.Parameter]:
    body = model.body

    layers = []
    layers += [nn.Sequential(body.conv1, body.bn1)]
    layers += [getattr(body, l) for l in list(body) if l.startswith("layer")]
    layers += [model.fpn]

    _param_groups = [list(layer.parameters()) for layer in layers]
    check_all_model_params_in_groups2(model, _param_groups)

    return _param_groups


def patch_param_groups(model: nn.Module) -> None:
    model.param_groups = MethodType(param_groups, model)



__all__ = [
    "resnet50_caffe_fpn_1x",
    "resnet50_fpn_1x",
    "resnet50_fpn_2x",
    "resnet101_caffe_fpn_1x",
    "resnet101_fpn_1x",
    "resnet101_fpn_2x",
    "resnext101_32x4d_fpn_1x",
    "resnext101_32x4d_fpn_2x",
    "resnext101_64x4d_fpn_1x",
    "resnext101_64x4d_fpn_2x",
]

from icevision.imports import *
from icevision.models.mmdet.utils import *


class MaskRCNNMMDetBackboneConfig(MMDetBackboneConfig):
    def __init__(self, **kwargs):
        super().__init__(model_name="mask_rcnn", **kwargs)


base_config_path = mmdet_configs_path / "mask_rcnn"
base_weights_url = "http://download.openmmlab.com/mmdetection/v2.0/mask_rcnn"

resnet50_caffe_fpn_1x = MaskRCNNMMDetBackboneConfig(
    config_path=base_config_path / "mask_rcnn_r50_caffe_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/mask_rcnn_r50_caffe_fpn_1x_coco/mask_rcnn_r50_caffe_fpn_1x_coco_bbox_mAP-0.38__segm_mAP-0.344_20200504_231812-0ebd1859.pth",
)

resnet50_fpn_1x = MaskRCNNMMDetBackboneConfig(
    config_path=base_config_path / "mask_rcnn_r50_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/mask_rcnn_r50_fpn_1x_coco/mask_rcnn_r50_fpn_1x_coco_20200205-d4b0c5d6.pth",
)

resnet50_fpn_2x = MaskRCNNMMDetBackboneConfig(
    config_path=base_config_path / "mask_rcnn_r50_fpn_2x_coco.py",
    weights_url=f"{base_weights_url}/mask_rcnn_r50_fpn_2x_coco/mask_rcnn_r50_fpn_2x_coco_bbox_mAP-0.392__segm_mAP-0.354_20200505_003907-3e542a40.pth",
)

resnet101_caffe_fpn_1x = MaskRCNNMMDetBackboneConfig(
    config_path=base_config_path / "mask_rcnn_r101_caffe_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/mask_rcnn_r101_caffe_fpn_1x_coco/mask_rcnn_r101_caffe_fpn_1x_coco_20200601_095758-805e06c1.pth",
)

resnet101_fpn_1x = MaskRCNNMMDetBackboneConfig(
    config_path=base_config_path / "mask_rcnn_r101_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/mask_rcnn_r101_fpn_1x_coco/mask_rcnn_r101_fpn_1x_coco_20200204-1efe0ed5.pth",
)

resnet101_fpn_2x = MaskRCNNMMDetBackboneConfig(
    config_path=base_config_path / "mask_rcnn_r101_fpn_2x_coco.py",
    weights_url=f"{base_weights_url}/mask_rcnn_r101_fpn_2x_coco/mask_rcnn_r101_fpn_2x_coco_bbox_mAP-0.408__segm_mAP-0.366_20200505_071027-14b391c7.pth",
)

resnext101_32x4d_fpn_1x = MaskRCNNMMDetBackboneConfig(
    config_path=base_config_path / "mask_rcnn_x101_32x4d_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/mask_rcnn_x101_32x4d_fpn_1x_coco/mask_rcnn_x101_32x4d_fpn_1x_coco_20200205-478d0b67.pth",
)

resnext101_32x4d_fpn_2x = MaskRCNNMMDetBackboneConfig(
    config_path=base_config_path / "mask_rcnn_x101_32x4d_fpn_2x_coco.py",
    weights_url=f"{base_weights_url}/mask_rcnn_x101_32x4d_fpn_2x_coco/mask_rcnn_x101_32x4d_fpn_2x_coco_bbox_mAP-0.422__segm_mAP-0.378_20200506_004702-faef898c.pth",
)

resnext101_64x4d_fpn_1x = MaskRCNNMMDetBackboneConfig(
    config_path=base_config_path / "mask_rcnn_x101_64x4d_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/mask_rcnn_x101_64x4d_fpn_1x_coco/mask_rcnn_x101_64x4d_fpn_1x_coco_20200201-9352eb0d.pth",
)

resnext101_64x4d_fpn_2x = MaskRCNNMMDetBackboneConfig(
    config_path=base_config_path / "mask_rcnn_x101_64x4d_fpn_2x_coco.py",
    weights_url=f"{base_weights_url}/mask_rcnn_x101_64x4d_fpn_2x_coco/mask_rcnn_x101_64x4d_fpn_2x_coco_20200509_224208-39d6f70c.pth",
)



# The draw functions are based from:
# https://github.com/fizyr/keras-retinanet/blob/master/keras_retinanet/utils/visualization.py
# https://github.com/fizyr/keras-maskrcnn/blob/master/keras_maskrcnn/utils/visualization.py

__all__ = [
    "draw_sample",
    "draw_record",
    "draw_pred",
    "draw_bbox",
    "draw_mask",
    "draw_keypoints",
]

from icevision.imports import *
from icevision.data import *
from icevision.core import *
from icevision.visualize.utils import *

# This should probably move elsewhere
from PIL import Image, ImageFont, ImageDraw
import PIL

DEFAULT_FONT_PATH = get_default_font()


def draw_sample(
    sample,
    class_map: Optional[ClassMap] = None,
    denormalize_fn: Optional[callable] = None,
    display_label: bool = True,
    display_bbox: bool = True,
    display_score: bool = True,
    display_mask: bool = True,
    display_keypoints: bool = True,
    font_path: Optional[os.PathLike] = DEFAULT_FONT_PATH,
    font_size: Union[int, float] = 12,
    label_color: Union[np.array, list, tuple, str] = "#C4C4C4",  # Mild Gray
    label_border_color: Union[np.array, list, tuple, str] = "#020303",  # Black,
    label_thin_border: bool = True,
    label_pad_width_factor: float = 0.02,
    label_pad_height_factor: float = 0.005,
    mask_blend: float = 0.5,
    mask_border_thickness: int = 7,
    color_map: Optional[dict] = None,  # label -> color mapping
    prettify: bool = True,
    prettify_func: Callable = str.capitalize,
    return_as_pil_img=False,
    # Args for plotting specific labels
    exclude_labels: List[str] = [],
    include_only: List[str] = None,
) -> Union[np.ndarray, PIL.Image.Image]:
    """
    Selected kwargs:

    * label_color: A <collection> of RGB values or a hex code string that defines
                   the color of all the plotted labels
    * label_border_color: Color of the border around the label
    * label_thin_border: Apply a thin border around the label. If false, applies
                         a thick border. If None, applies no border
    * label_pad_width_factor: Amount of padding to apply relative to the image's width.
                              Applies padding to bbox coords if padding bbox
                              labels else to the top-left of the image for classif labels
    * mask_blend: Degree of transparency of the mask. 1 = opaque, 0 = transparent
    * mask_border_thickness: Degree of thickness of the mask. Must be an odd number
    * color_map: An optional dictionary that maps the label => color-value
    * prettify: Format labels based on `prettify_func`
    * prettify_func: A string -> string processing function
    * return_as_pil_image: If true, returns the sample as a PIL image, else np.array
    * exclude_labels: (Optional) List of labels that you'd like to exclude from being plotted
    * include_only: (Optional) List of labels that must be exclusively plotted. Takes
                    precedence over `exclude_labels` (?)
    """
    img = sample.img.copy()

    if denormalize_fn is not None:
        img = denormalize_fn(img)

    for task, composite in sample.task_composites.items():
        # Should break if no ClassMap found in composite.
        #  Should be as the only composite without ClassMap should be
        #  `sample.common`. This is a foundational assumption? #NOTE
        class_map = getattr(composite, "class_map", None)
        for label, bbox, mask, keypoints, score in itertools.zip_longest(
            getattr(composite, "labels", []),  # list of strings
            getattr(composite, "bboxes", []),
            getattr(composite, "masks", []),
            getattr(composite, "keypoints", []),
            getattr(composite, "scores", []),
        ):
            # random color by default
            color = (np.random.random(3) * 0.6 + 0.4) * 255

            # logic for plotting specific labels only
            # `include_only` > `exclude_labels`
            if not label == []:
                label_str = (
                    class_map.get_by_name(label) if class_map is not None else ""
                )
                if include_only is not None:
                    if not label_str in include_only:
                        continue
                elif label_str in exclude_labels:
                    continue

            # if color-map is given and `labels` are predicted
            # then set color accordingly
            if color_map is not None:
                color = np.array(color_map[label_str]).astype(np.float)

            if display_mask and mask is not None:
                img = draw_mask(
                    img=img,
                    mask=mask,
                    color=color,
                    blend=mask_blend,
                    border_thickness=mask_border_thickness,
                )
            if display_bbox and bbox is not None:
                img = draw_bbox(img=img, bbox=bbox, color=color)
            if display_keypoints and keypoints is not None:
                img = draw_keypoints(img=img, kps=keypoints, color=color)
            if display_label and label is not None:
                img = draw_label(
                    img=img,
                    label=label,
                    score=score if display_score else None,
                    bbox=bbox,
                    mask=mask,
                    class_map=class_map,
                    color=label_color,
                    border_color=label_border_color,
                    pad_width_factor=label_pad_width_factor,
                    pad_height_factor=label_pad_height_factor,
                    thin_border=label_thin_border,
                    font_size=font_size,
                    font=font_path,
                    prettify=prettify,
                    prettify_func=prettify_func,
                    return_as_pil_img=False,  # should this always be False??
                )
    if return_as_pil_img:
        # may or may not be a PIL Image based on `display_label`
        return img if isinstance(img, PIL.Image.Image) else PIL.Image.fromarray(img)
    else:
        # will be a `np.ndarray` by default so no need for casting
        return img


def draw_label(
    img: np.ndarray,
    label: int,
    score: Optional[float],
    color: Union[np.ndarray, list, tuple],
    border_color: Union[np.ndarray, list, tuple],
    class_map: Optional[ClassMap] = None,
    bbox=None,
    mask=None,
    font: Union[int, os.PathLike, None] = None,
    font_size: Union[int, float] = 12,
    prettify: bool = True,
    prettify_func: Callable = str.capitalize,
    return_as_pil_img=False,
    pad_width_factor=0.02,
    pad_height_factor=0.005,
    thin_border=True,
) -> Union[np.ndarray, PIL.Image.Image]:
    # finds label position based on bbox or mask
    if bbox is not None:
        x, y, _, _ = bbox.xyxy
    elif mask is not None:
        y, x = np.unravel_index(mask.data.argmax(), mask.data.shape)
    else:
        x, y = 0, 0

    if class_map is not None:
        if isinstance(label, int):
            # TODO: This may never get triggered because we're looping
            # over composite.labels which is a list of strings
            caption = class_map.get_by_id(label)
        else:
            caption = label
    else:
        caption = str(label)
    if prettify:
        # We could introduce a callback here for more complex label renaming
        caption = prettify_func(caption)

    # Append label confidence to caption if applicable
    if score is not None:
        if prettify:
            score = f"{score * 100: .2f}%"
        caption = f"{caption}: {score}"

    if not Path(font).exists():
        # PIL throws cryptic errors for wrong filepaths, so let's catch it earlier here
        raise FileNotFoundError(f"{font} file doesn't exist")
    return _draw_label(
        img=img,
        caption=caption,
        x=x,
        y=y,
        color=color,
        border_color=border_color,
        font_path=font,
        font_size=int(font_size),
        return_as_pil_img=return_as_pil_img,
        pad_width_factor=pad_width_factor,
        pad_height_factor=pad_height_factor,
        thin_border=thin_border,
    )


def _draw_label(
    img: np.ndarray,
    caption: str,
    x: int,
    y: int,
    color: Union[np.ndarray, list, tuple],
    border_color: Union[np.ndarray, list, tuple],
    font_path=DEFAULT_FONT_PATH,
    font_size: int = 20,
    return_as_pil_img: bool = False,
    pad_width_factor=0.02,
    pad_height_factor=0.005,
    thin_border=True,
) -> Union[PIL.Image.Image, np.ndarray]:
    """Draw labels on the image"""
    font = PIL.ImageFont.truetype(font_path, size=font_size)
    color = as_rgb_tuple(color)
    border_color = as_rgb_tuple(border_color)

    height, width = img.shape[:2]
    x_pad = height * pad_width_factor
    y_pad = width * pad_height_factor
    x, y = x + x_pad, y + y_pad

    img = PIL.Image.fromarray(img)
    draw = ImageDraw.Draw(img)

    if thin_border is not None:
        # Draw thin / thick border around text
        draw.text(
            (x - 1, y if thin_border else y - 1), caption, font=font, fill=border_color
        )
        draw.text(
            (x + 1, y if thin_border else y - 1), caption, font=font, fill=border_color
        )
        draw.text(
            (x if thin_border else x - 1, y - 1), caption, font=font, fill=border_color
        )
        draw.text(
            (x if thin_border else x + 1, y + 1), caption, font=font, fill=border_color
        )

    # Now draw text over the border
    draw.text((x, y), caption, font=font, fill=color)
    if return_as_pil_img:
        return img
    else:
        return np.array(img)


def draw_record(
    record,
    class_map: Optional[ClassMap] = None,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    display_score: bool = True,
    display_keypoints: bool = True,
    font_path: Optional[os.PathLike] = DEFAULT_FONT_PATH,
    font_size: Union[int, float] = 12,
    label_color: Union[np.array, list, tuple, str] = "#C4C4C4",  # Mild Gray
    mask_blend: float = 0.5,
    mask_border_thickness: int = 7,
    color_map: Optional[dict] = None,  # label -> color mapping
    prettify: bool = True,
    prettify_func: Callable = str.capitalize,
    return_as_pil_img=False,
    # Args for plotting specific labels
    exclude_labels: List[str] = [],
    include_only: List[str] = None,
):
    sample = record.load()
    return draw_sample(
        sample=sample,
        class_map=class_map,
        display_label=display_label,
        display_score=display_score,
        display_bbox=display_bbox,
        display_mask=display_mask,
        display_keypoints=display_keypoints,
        font_path=font_path,
        font_size=font_size,
        label_color=label_color,
        mask_blend=mask_blend,
        mask_border_thickness=mask_border_thickness,
        color_map=color_map,
        prettify=prettify,
        prettify_func=prettify_func,
        return_as_pil_img=return_as_pil_img,
        exclude_labels=exclude_labels,
        include_only=include_only,
    )


def draw_pred(
    pred: Prediction,
    denormalize_fn: Optional[callable] = None,
    display_label: bool = True,
    display_score: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    font_path: Optional[os.PathLike] = DEFAULT_FONT_PATH,
    font_size: Union[int, float] = 12,
    label_color: Union[np.array, list, tuple, str] = "#C4C4C4",  # Mild Gray
    mask_blend: float = 0.5,
    mask_border_thickness: int = 7,
    color_map: Optional[dict] = None,  # label -> color mapping
    prettify: bool = True,
    prettify_func: Callable = str.capitalize,
    return_as_pil_img=False,
    # Args for plotting specific labels
    exclude_labels: List[str] = [],
    include_only: List[str] = None,
):
    return draw_sample(
        sample=pred.pred,
        denormalize_fn=denormalize_fn,
        display_label=display_label,
        display_score=display_score,
        display_bbox=display_bbox,
        display_mask=display_mask,
        font_path=font_path,
        font_size=font_size,
        label_color=label_color,
        mask_blend=mask_blend,
        mask_border_thickness=mask_border_thickness,
        color_map=color_map,
        prettify=prettify,
        prettify_func=prettify_func,
        return_as_pil_img=return_as_pil_img,
        exclude_labels=exclude_labels,
        include_only=include_only,
    )


def draw_bbox(
    img: np.ndarray,
    bbox: BBox,
    color: Tuple[int, int, int],
    gap: bool = True,
):
    """Draws a box on an image with a given color.
    # Arguments
        image     : The image to draw on.
        box       : A list of 4 elements (x1, y1, x2, y2).
        color     : The color of the box.
    """

    # Calculate image dimensions
    dims = sorted(img.shape, reverse=True)
    color = as_rgb_tuple(color)
    img = PIL.Image.fromarray(img)
    draw = PIL.ImageDraw.Draw(img)

    # corner thickness is linearly correlated with the smaller image dimension.
    # We use the smaller image dimension rather than image area so as to avoid
    # overly thick lines for large non-square images prior to transforming
    # images. We set lower and upper bounds for corner thickness.
    min_corner = 1
    max_corner = 15
    corner_thickness = int(0.005 * dims[1] + min_corner)
    if corner_thickness > max_corner:
        corner_thickness = int(max_corner)

    corner_length = int(0.021 * dims[1] + 2.25)

    # inner thickness of bboxes with corners
    inner_thickness = int(1 + 0.0005 * dims[1])

    # bbox thickness of bboxes without corners
    min_bbox = 1
    max_bbox = 8
    bbox_thickness = int(0.0041 * dims[1] - 0.0058)
    if bbox_thickness < min_bbox:
        bbox_thickness = min_bbox
    if bbox_thickness > max_bbox:
        bbox_thickness = int(max_bbox)

    if gap == False:
        xyxy = tuple(np.array(bbox.xyxy, dtype=int))
        draw.rectangle(xyxy, fill=None, outline=color, width=bbox_thickness)
        return np.array(img)

    xmin, ymin, xmax, ymax = tuple(np.array(bbox.xyxy, dtype=int))

    points = [0] * 12
    points[0] = (xmin, ymin + corner_length)
    points[1] = (xmin, ymin)
    points[2] = (xmin + corner_length, ymin)

    points[3] = (xmax - corner_length, ymin)
    points[4] = (xmax, ymin)
    points[5] = (xmax, ymin + corner_length)

    points[6] = (xmax, ymax - corner_length)
    points[7] = (xmax, ymax)
    points[8] = (xmax - corner_length, ymax)

    points[9] = (xmin + corner_length, ymax)
    points[10] = (xmin, ymax)
    points[11] = (xmin, ymax - corner_length)

    if (
        ymax - (ymin + 4 * corner_length) < corner_length
        or xmax - (xmin + 4 * corner_length) < corner_length
    ):
        for i in range(4):
            draw.line(
                xy=(points[i * 3 + 1], points[10 - 3 * i]),
                fill=color,
                width=bbox_thickness,
            )
        for i in range(2):
            draw.line(
                xy=(points[6 * i + 1], points[i * 6 + 4]),
                fill=color,
                width=bbox_thickness,
            )
    else:
        for i in range(2):
            for j in range(2):
                draw.line(
                    xy=(points[i * 6 + j * 4], points[i * 6 + j * 4 + 1]),
                    fill=color,
                    width=corner_thickness,
                    joint=None,
                )
                draw.line(
                    xy=(points[i * 3 + 1], points[10 - 3 * i]),
                    fill=color,
                    width=inner_thickness,
                    joint=None,
                )
        for i in range(2):
            for j in range(2):
                draw.line(
                    xy=(points[i * 6 + j * 2 + 1], points[i * 6 + j * 2 + 2]),
                    fill=color,
                    width=corner_thickness,
                    joint=None,
                )
                draw.line(
                    xy=(points[6 * i + 1], points[i * 6 + 4]),
                    fill=color,
                    width=inner_thickness,
                    joint=None,
                )

    return np.array(img)


def draw_mask(
    img: np.ndarray,
    mask: MaskArray,
    color: Tuple[int, int, int],
    blend: float = 0.5,
    border_thickness: int = 7,
):
    # Border thickness must be an odd integer
    if border_thickness % 2 == 0:
        # TODO: Shall we throw an error, or change the value to the nearest
        # even integer automatically and raise a warning?
        raise ValueError(
            f"`border_thickness` must be an odd number. You entered {border_thickness}"
        )
    img = PIL.Image.fromarray(img)
    w, h = img.size

    mask_idxs = np.where(mask.data)

    # Add alpha with 0 transparency. We draw the mask with border color first
    color = np.append(color, 255)
    mask_arr = np.zeros((h, w, 4), dtype=np.uint8)
    mask_arr[mask_idxs] = color

    # Now create a second mask and draw the desired color on top of the
    # border mask. If `border_thickness` is 0, this replaces the border mask
    _mask = Image.fromarray(mask_arr)
    _mask = _mask.filter(PIL.ImageFilter.MinFilter(border_thickness))
    _mask_idx = np.where(_mask.convert("L", palette=Image.ADAPTIVE))
    mask_arr[_mask_idx] = np.append(color[:3], blend * 255)

    # Create RGBA PIL mask image
    mask_pil = PIL.Image.fromarray(mask_arr, mode="RGBA")

    # Blend everything keeping the alpha
    # Key concept is that alpha for non-mask pixels are 0 (transparent)
    img.putalpha(255)
    img = PIL.Image.alpha_composite(img, mask_pil)

    return np.array(img)


def draw_keypoints(
    img: np.ndarray,
    kps: KeyPoints,
    color: Tuple[int, int, int],
):
    x, y, v = kps.x, kps.y, kps.visible

    # calculate scaling for points and connections
    img_h, img_w, _ = img.shape
    img_area = img_h * img_w
    img = PIL.Image.fromarray(img)
    draw = PIL.ImageDraw.Draw(img)
    dynamic_size = int(0.01867599 * (img_area ** 0.4422045))
    dynamic_size = max(dynamic_size, 1)

    # draw connections
    if kps.metadata is not None and kps.metadata.connections is not None:
        for connection in kps.metadata.connections:
            if v[connection.p1] > 0 and v[connection.p2] > 0:
                draw.line(
                    xy=(
                        (int(x[connection.p1]), int(y[connection.p1])),
                        (int(x[connection.p2]), int(y[connection.p2])),
                    ),
                    fill=connection.color,
                    width=dynamic_size,
                )

    # draw points
    for x_c, y_c in zip(x[v > 0], y[v > 0]):
        radius = dynamic_size
        x1 = x_c - radius
        x2 = x_c + radius
        y1 = y_c - radius
        y2 = y_c + radius
        draw.ellipse(
            xy=[x1, y1, x2, y2],
            # xy=[x1, x0, y1, y0],
            fill=as_rgb_tuple(color),
            outline=None,
            width=0,
        )

    return np.array(img)



__all__ = ["sum_losses_mmdet", "loop_mmdet"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.data import *
from icevision.models.interpretation import _move_to_device
from icevision.core.record_components import LossesRecordComponent


def sum_losses_mmdet(losses_dict):
    loss_ = {}
    for k, v in losses_dict.items():
        if k.startswith("loss"):
            if isinstance(v, torch.Tensor):
                loss_[k] = float(v.cpu().numpy())
            elif isinstance(v, list):
                loss_[k] = sum([float(o.cpu().numpy()) for o in v])

    loss_["loss_total"] = sum(loss_.values())
    return loss_


def loop_mmdet(dl, model, losses_stats, device):
    samples_plus_losses = []

    with torch.no_grad():
        for data, sample in pbar(dl):
            torch.manual_seed(0)
            _, data = _move_to_device(None, data, device)
            loss = model(**data)
            loss = sum_losses_mmdet(loss)

            for l in losses_stats.keys():
                losses_stats[l].append(loss[l])

            loss_comp = LossesRecordComponent()
            loss_comp.set_losses(loss)
            sample[0].add_component(loss_comp)
            samples_plus_losses.append(sample[0])

    return samples_plus_losses, losses_stats



__all__ = ["build_model"]

from icevision.imports import *
from mmcv import Config
from mmdet.models import build_detector
from mmcv.runner import load_checkpoint
from icevision.models.mmdet.utils import *


def build_model(
    model_type: str,
    backbone: MMDetBackboneConfig,
    num_classes: int,
    pretrained: bool = True,
    checkpoints_path: Optional[Union[str, Path]] = "checkpoints",
    force_download=False,
) -> nn.Module:

    cfg, weights_path = create_model_config(
        backbone=backbone,
        pretrained=pretrained,
        checkpoints_path=checkpoints_path,
        force_download=force_download,
    )

    if model_type == "one_stage_detector_bbox":
        cfg.model.bbox_head.num_classes = num_classes - 1

    if model_type == "two_stage_detector_bbox":
        # Sparse-RCNN has a list of bbox_head whereas Faster-RCNN has only one
        if isinstance(cfg.model.roi_head.bbox_head, list):
            for bbox_head in cfg.model.roi_head.bbox_head:
                bbox_head["num_classes"] = num_classes - 1
        else:
            cfg.model.roi_head.bbox_head.num_classes = num_classes - 1

    if model_type == "two_stage_detector_mask":
        cfg.model.roi_head.bbox_head.num_classes = num_classes - 1
        cfg.model.roi_head.mask_head.num_classes = num_classes - 1

    if (pretrained == False) or (weights_path is not None):
        cfg.model.pretrained = None

    _model = build_detector(cfg.model, cfg.get("train_cfg"), cfg.get("test_cfg"))

    if pretrained and (weights_path is not None):
        load_checkpoint(_model, str(weights_path))

    _model.param_groups = MethodType(param_groups, _model)

    return _model



from icevision.all import *

data_dir = icedata.fridge.load_data()

parser = icedata.fridge.parser(data_dir)

train_records, valid_records = parser.parse(autofix=False)

presize, size = 256, 128

train_tfms = tfms.A.Adapter(
    [*tfms.A.aug_tfms(size=size, presize=presize), tfms.A.Normalize()]
)
valid_tfms = tfms.A.Adapter([*tfms.A.resize_and_pad(size=size), tfms.A.Normalize()])

train_ds = Dataset(train_records, train_tfms)
valid_ds = Dataset(valid_records, valid_tfms)

model_type = models.mmdet.faster_rcnn
train_dl = model_type.train_dl(train_ds, batch_size=2, shuffle=True)
valid_dl = model_type.valid_dl(valid_ds, batch_size=2, shuffle=False)

### MODEL

from mmcv import Config

cfg = Config.fromfile(
    "~/git/mmdetection/configs/faster_rcnn/faster_rcnn_r50_fpn_1x_coco.py"
)
cfg.model.roi_head.bbox_head.num_classes = len(parser.class_map) - 1

from mmdet.models import build_detector

model = build_detector(cfg.model, cfg.get("train_cfg"), cfg.get("test_cfg"))


class LitModel(model_type.lightning.ModelAdapter):
    def configure_optimizers(self):
        return SGD(self.parameters(), lr=1e-3)


lit_model = LitModel(model)
trainer = pl.Trainer(
    max_epochs=10, gpus=1, num_sanity_val_steps=0, check_val_every_n_epoch=1
)
trainer.fit(lit_model, train_dl, valid_dl)



__all__ = ["Prediction"]

from icevision.imports import *
from icevision.core import *


class Prediction:
    def __init__(self, pred: BaseRecord, ground_truth: Optional[BaseRecord] = None):
        self.ground_truth = ground_truth
        self.pred = pred

        # TODO: record_id, img_size and stuff has to be set even if ground_truth is not stored
        if ground_truth is not None:
            pred.set_record_id(ground_truth.record_id)
            pred.set_img_size(ground_truth.original_img_size, original=True)
            pred.set_img_size(ground_truth.img_size)
            # HACK
            if ground_truth.img is not None:
                pred.set_img(ground_truth.img)

    def __getattr__(self, name):
        if name == "pred":
            raise AttributeError

        return getattr(self.pred, name)



__all__ = ["display_model_choice_ui", "lib_info", "get_model_info"]


from icevision import models
import ipywidgets as widgets
from IPython.display import display


lib_info = {
    "lib_type": None,
    "model_type": None,
    "backbone_type": None,
    "backbone": None,
    "model_list": [],
    "backbone_list": [],
}


def reset_lib_info():
    lib_info.update({"model_list": []})
    lib_info.update({"backbone_list": []})
    lib_info.update({"lib_type": None})
    lib_info.update({"model_type": None})
    lib_info.update({"backbone_type": None})
    lib_info.update({"backbone": None})


def get_model_info():
    model_type = lib_info["model_type"]
    backbone = lib_info["backbone"]

    return model_type, backbone


# Creating dropdown widgets
libraries_available = widgets.Dropdown(
    options=["MMDetection", "Ross Wightman", "Torchvision"],
    # options=[models.ross, models.torchvision],
    description="Libraries",
    disabled=False,
)

models_available = widgets.Dropdown(
    options=[""],
    description="Models",
    disabled=False,
)

backbones_available = widgets.Dropdown(
    options=[""],
    description="Backbones",
    disabled=False,
)


def display_model_choice_ui():
    # Observe dropdown widget changes
    libraries_available.observe(od_library_change, names="value")
    models_available.observe(od_model_change, names="value")
    backbones_available.observe(od_backbone_change, names="value")

    # display dropdown widgets
    display(libraries_available)
    display(models_available)
    display(backbones_available)


def od_library_change(change):
    reset_lib_info()
    lib_name = change.new

    if lib_name == "Torchvision":
        lib_type = models.torchvision
        model_list = ["faster_rcnn", "retinanet", "mask_rcnn", "keypoint_rcnn"]

    if lib_name == "MMDetection":
        lib_type = models.mmdet.models
        model_list = ["sparse_rcnn", "retinanet", "mask_rcnn"]

    if lib_name == "Ross Wightman":
        lib_type = models.ross
        model_list = ["efficientdet"]

    lib_info.update({"lib_type": lib_type})
    lib_info.update({"model_list": model_list})

    models_available.options = model_list


def od_model_change(change):
    model_name = change.new
    lib_type = lib_info["lib_type"]
    model_type = getattr(lib_type, model_name)
    backbone_type = getattr(model_type, "backbones")
    backbone_list = [item for item in dir(model_type.backbones) if "__" not in item]

    lib_info.update({"model_type": model_type})
    lib_info.update({"backbone_type": backbone_type})
    lib_info.update({"backbone_list": backbone_list})
    backbones_available.options = backbone_list


def od_backbone_change(change):
    backbone_name = change.new
    model_type = lib_info["model_type"]
    backbone_type = lib_info["backbone_type"]

    backbone = getattr(backbone_type, backbone_name)

    lib_info.update({"backbone": backbone})



__all__ = [
    "predict",
    "predict_dl",
    "convert_raw_prediction",
    "convert_raw_predictions",
]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.models.utils import _predict_dl
from icevision.data import *
from icevision.models.torchvision.faster_rcnn.dataloaders import *


@torch.no_grad()
def _predict_batch(
    model: nn.Module,
    batch: Sequence[torch.Tensor],
    records: Sequence[BaseRecord],
    detection_threshold: float = 0.5,
    device: Optional[torch.device] = None,
):
    model.eval()
    device = device or model_device(model)
    batch = [o.to(device) for o in batch]

    raw_preds = model(*batch)
    return convert_raw_predictions(
        raw_preds=raw_preds, records=records, detection_threshold=detection_threshold
    )


def predict(
    model: nn.Module,
    dataset: Dataset,
    detection_threshold: float = 0.5,
    device: Optional[torch.device] = None,
) -> List[Prediction]:
    batch, records = build_infer_batch(dataset)
    return _predict_batch(
        model=model,
        batch=batch,
        records=records,
        detection_threshold=detection_threshold,
        device=device,
    )


def predict_dl(
    model: nn.Module,
    infer_dl: DataLoader,
    show_pbar: bool = True,
    **predict_kwargs,
):
    return _predict_dl(
        predict_fn=_predict_batch,
        model=model,
        infer_dl=infer_dl,
        show_pbar=show_pbar,
        **predict_kwargs,
    )


def convert_raw_predictions(
    raw_preds, records: Sequence[BaseRecord], detection_threshold: float
):
    return [
        convert_raw_prediction(
            raw_pred=raw_pred,
            record=record,
            detection_threshold=detection_threshold,
        )
        for raw_pred, record in zip(raw_preds, records)
    ]


def convert_raw_prediction(
    raw_pred: dict, record: BaseRecord, detection_threshold: float
):
    above_threshold = raw_pred["scores"] >= detection_threshold

    labels = raw_pred["labels"][above_threshold]
    labels = labels.detach().cpu().numpy()

    scores = raw_pred["scores"][above_threshold]
    scores = scores.detach().cpu().numpy()

    boxes = raw_pred["boxes"][above_threshold]
    bboxes = []
    for box_tensor in boxes:
        xyxy = box_tensor.cpu().numpy()
        bbox = BBox.from_xyxy(*xyxy)
        bboxes.append(bbox)

    pred = BaseRecord(
        (
            ScoresRecordComponent(),
            ImageRecordComponent(),
            InstancesLabelsRecordComponent(),
            BBoxesRecordComponent(),
        )
    )
    pred.detection.set_class_map(record.detection.class_map)
    pred.detection.set_scores(scores)
    pred.detection.set_labels_by_id(labels)
    pred.detection.set_bboxes(bboxes)
    pred.detection.above_threshold = above_threshold

    return Prediction(pred=pred, ground_truth=record)
